import os.path
from typing import Tuple

import PIL
import imageio
import numpy as np

# ------------------------------------------------------------------------------
#           load_lunarimage
# ------------------------------------------------------------------------------
from altius_l1pp_inflight.configuration import altius_config
from altius_l1pp_inflight.psf.psf_interface import PsfSimpleGaussian


def load_normalized_lunarimage(required_size=None) -> np.ndarray:
    if required_size is None:
        required_size = (1500 // 4, 1500 // 4)

    base, name = os.path.split(__file__)
    filename = os.path.join(base, 'lunarmaps', 'fullmoon.png')
    fd = np.array(imageio.imread(filename), dtype=float)
    rawdata = np.sqrt(np.sum(fd[:, :, 0:2] * fd[:, :, 0:2], axis=2)) / 756.0
    largemoon = PIL.Image.fromarray(rawdata)
    moon = np.array(largemoon.resize(required_size, resample=PIL.Image.LANCZOS))
    moon *= 1.0 / np.max(moon)
    return moon


# ------------------------------------------------------------------------------
#           class lunar_psf_simulator
# ------------------------------------------------------------------------------

class lunar_psf_simulator:

    def __init__(self, config_file=None, position=(250, 250), channel='uv', mode='native'):
        self._config_file = config_file
        self._channel = channel
        self._mode = mode
        self._psf = None
        self._moon = None
        self.mooncenter = None
        self.moonimage = None
        self.moonbounds = None
        self.create_lunar_image(position)

    # ------------------------------------------------------------------------------
    #           create_lunar_image
    # ------------------------------------------------------------------------------

    def create_lunar_image(self, position: Tuple[int, int], counts_per_raw_pixel=10000.0):
        """
        Creates a detector image correspondinhg to the given instrument. It then places an image of the full moon
        that is 1/4 of the  x dimensions
        """
        ny, nx = altius_config.get_detector_pixel_dimensions(self._channel, self._mode, self._config_file)
        ybin, xbin = altius_config.get_binning(self._config_file, self._channel)
        self._moon = load_normalized_lunarimage(required_size=(ny // 4, nx // 4))
        centrepos = np.array(position, dtype='int32')  # get the location where the user wants the centyer of the moon
        moondiameter = np.array(self._moon.shape,
                                dtype='int32')  # Get the diameter of the moon image. This is just the width of the picture
        moonradius = moondiameter // 2  # Get radius of lunar image in both dimensions
        llmoon = np.array((0, 0),
                          dtype='int32')  # Lower left corner  of used moon coordinate on image (maybe off the edge)
        urmoon = moondiameter  # Upper right corner of used moon coordinates on the image (maybe off an edge)
        llimage = centrepos - moonradius  # Get the lower left corner of the moon on the image
        urimage = llimage + moondiameter  # get the upper right corner of the moon on the image
        imagesize = np.array((ny, nx), dtype='int32')  # get the size of the image in a an array

        for i in range(2):  # Check that moon image is not outside bounds of final image. If it is then truncate
            if llimage[i] < 0:  # If the left/bottom edge of the moon is off the left/bottom side
                llmoon[i] = np.abs(llimage[i])  # then truncate the amount of moon image actually mapped
                llimage[i] = 0  # and set the start point
            if urimage[i] > imagesize[i]:  # similarly with the upper bound
                urmoon[i] = urmoon[i] - (urimage[i] - imagesize[i])  # If it extends over the bound
                urimage[i] = imagesize[i]
        factor = xbin * ybin * counts_per_raw_pixel  # Get the factor for the image pixel so we get the right counts
        self.moonimage = np.zeros(imagesize)
        self.moonimage[llimage[0]:urimage[0], llimage[1]:urimage[1]] = self._moon[llmoon[0]:urmoon[0],
                                                                                  llmoon[1]:urmoon[1]] * factor
        self.mooncenter = centrepos
        self.moonbounds = np.array([llimage, urimage])

    # ------------------------------------------------------------------------------
    #           +psf
    # ------------------------------------------------------------------------------
    @property
    def psf(self):
        if self._psf is None:
            # self._psf = ALTIUS_PSF(self._inst)
            self._psf = PsfSimpleGaussian(3.5)  # TODO this will pull in real data
        return self._psf

    # ------------------------------------------------------------------------------
    #           get_radial_array
    # ------------------------------------------------------------------------------

    def get_radial_points_array_centered_on_moon(self):

        xp = np.arange(0, self.moonimage.shape[1]).reshape([1, self.moonimage.shape[1]]) - self.mooncenter[
            1]  # Get the range of X coordinates and resize so they are appropriate for array telegraphing
        yp = np.arange(0, self.moonimage.shape[0]).reshape([self.moonimage.shape[0], 1]) - self.mooncenter[0]
        R = np.sqrt(yp * yp + xp * xp)
        IR = np.array(R + 0.5, dtype='int32')
        Nr = np.max(IR) + 1
        Npts = np.zeros([Nr])
        for i in range(Nr):
            Npts[i] = np.count_nonzero(IR == i)
        return Npts

    # ------------------------------------------------------------------------------
    #           scattered_light_along_centerline
    # ------------------------------------------------------------------------------

    def scattered_light_along_radial_line(self):
        """
        Calculates the lunar scattered/PSF light along a radial line from the center of the moon.
        The returned signal has been multiplied by the number of points on the
        detector surface lying at each radius, (ie integrated in azimuth).

        returns
        -------
        A two element tuple. The first element is the radial `+psf signal` from the center of the moon
        out expressed as a 1 dimensional array. The second element is the number of points at each radial bin
        expressed as a 1 dimensional array

        """

        xp = np.arange(self.mooncenter[1], self.moonimage.shape[
            1])  # The coordinate of X along a horizontal row from the center to the rightmost edge
        yp = np.full_like(xp, self.mooncenter[
            0])  # The coordinate of Y along a horizontal row through the center o fthe moon
        signal = np.zeros(
            xp.shape)  # Array to hold the scattered signal along a horizontal row through the center of the moon

        for y0 in range(self.moonbounds[0, 0],
                        self.moonbounds[1, 0]):  # NOw raster scan across the image columns of the moon
            for x0 in range(self.moonbounds[0, 1],
                            self.moonbounds[1, 1]):  # and raster scan across the image rows of the moon
                deltax = xp - x0  # for each pixel on the moon, calculate the delta x to each element of the full horizontal row
                deltay = yp - y0  # and the delta y
                b = self.moonimage[y0, x0]  # get the value of the lunar image at this pixel
                p = self.psf.at_offset(deltay, deltax)  # get the PSF for each element acros sthe horizontal row
                signal += p * b  # and sum the signal
        N = self.get_radial_points_array_centered_on_moon()  # Get the radial distribution of points around the center of the moon
        N = N[0:signal.size]  # Adjust the radial profile accordingly
        signal *= N  # and multiply signal by the number of radial points around the image.
        return signal, N
