import numpy as np
from skimage.filters import sobel

SOLAR_RADIUS = 0.69634e6  # km
LUNAR_RADIUS = 1737.1  # km
D_1AU = 1.495978707e8  # km
EARTH_TO_MOON_DISTANCE_KM = 384400


class DiskImageGeometry:
    """
        Class for handling the pixel coordinates relative to the boresight of the
        instrument, and relative to the disk (sun, moon) center.  Includes calculations
        involving the solar disk parameter mu, and solid angles of pixels.

        This class is meant to be used for a constant size and positioned set of pixels, with the bottom
        left pixel corner as [0, 0], and half-pixels representing the center of each
        pixel.

        The basic constants are set, such as the body radius and detector width, and the position
        of the disk on the detector is toggled and recalculated by calling :code:`set_disk`
        for each new position.
    """

    def __init__(self,
                 horizontal_pixel_width: int,
                 vertical_pixel_width: int,
                 horz_fov_degrees: float,
                 vert_fov_degrees: float,
                 body_radius_km: float,
                 vertical_pixel_boresight: float,
                 horizontal_pixel_boresight: float,
                 padding: int = 0):
        """
        Parameters
        ----------
        horizontal_pixel_width : int
            Width of the image in pixels.

        vertical_pixel_width : int
            Height of the image in pixels.

        body_radius_km : float
            The physical radius of the (assumed) spherical body (sun, moon)

        horz_fov_degrees : float
            Horizontal field of view of the detector, in degrees,

        vert_fov_degrees : float
            Vertical field of view of the detector, in degrees

        vertical_pixel_boresight : float
            Vertical pixel coordinate of the boresight of the detector.

        horizontal_pixel_boresight : float
            Horizontal pixel coordinate of the boresight of the detector.

        padding : int
            Extra pixels to pad on the sides to handle when the disk is partially off the detector.
        """

        self._padding = padding
        self._horizontal_pixel_width = horizontal_pixel_width + 2 * padding
        self._vertical_pixel_width = vertical_pixel_width + 2 * padding
        self._vert_pixel_width_no_padding = vertical_pixel_width
        self._horz_pixel_width_no_padding = horizontal_pixel_width
        self._body_radius_km = body_radius_km
        self._vertical_pixel_boresight = vertical_pixel_boresight + padding
        self._horizontal_pixel_boresight = horizontal_pixel_boresight + padding
        self._horz_fov_deg = horz_fov_degrees * self._horizontal_pixel_width / self._horz_pixel_width_no_padding
        self._vert_fov_deg = vert_fov_degrees * self._vertical_pixel_width / self._vert_pixel_width_no_padding
        DH, DV = self._get_delta_pixels_relative_boresight()
        self._horizontal_pixel_mesh_rel_to_boresight = DH
        self._vertical_pixel_mesh_rel_to_boresight = DV
        self._delta_omega_view, self._theta_view, self._phi_view = self.compute_delta_solid_angles(
            self._horizontal_pixel_mesh_rel_to_boresight,
            self._vertical_pixel_mesh_rel_to_boresight)

    def get_delta_pixels_relative_to_boresight(self):
        """
        Returns
        -------
        horz_mesh, vert_mesh : np.ndarray, np.ndarray
            Mesh grid from calls to np.meshgrid of the pixel distances of each pixel relative to the boresight
        """
        return self._horizontal_pixel_mesh_rel_to_boresight, self._vertical_pixel_mesh_rel_to_boresight

    def get_delta_pixels_relative_disk_center(self):
        """
            Returns
            -------
            horz_mesh, vert_mesh : np.ndarray, np.ndarray
                Mesh grid from calls to np.meshgrid of the pixel distances of each pixel relative to the
                current disk center
        """
        return self._horizontal_pixel_mesh_relative_to_disk_center, self._vertical_pixel_mesh_relative_to_disk_center

    def with_padding(self, padding: float):
        """
        Returns an instance of disk geometry with extra padding on the sides

        Parameters
        ----------
        padding: float
            Multiplier for extra width of the detector

        Returns
        -------
            disk_geometry : DiskImageGeometry
        """
        pixel_padding = int(np.max([int(np.round(padding * self._horizontal_pixel_width) / 2),
                                    int(np.round(padding * self._vertical_pixel_width)) / 2]))
        return SolarGeometry(horizontal_pixel_width=self._horizontal_pixel_width,
                             vertical_pixel_width=self._vertical_pixel_width,
                             horz_fov_degrees=self._vert_fov_deg,
                             vert_fov_degrees=self._horz_fov_deg,
                             vertical_pixel_boresight=self._vertical_pixel_boresight,
                             horizontal_pixel_boresight=self._horizontal_pixel_boresight,
                             padding=pixel_padding)

    def get_image_bounds(self):
        vl = self._padding
        vu = self._padding + self._vert_pixel_width_no_padding
        hl = self._padding
        hu = self._padding + self._horz_pixel_width_no_padding
        return vl, vu, hl, hu

    def delta_omega_view(self, include_padding=False):
        """
            Parameters
            ----------
            include_padding: bool, Optional
                If true, return the array of the detector with padded pixels on the sides.

            Returns
            -------
            solid_angles : np.ndarray
                Detector-sized array of the solid angles of each pixel with the boresight of the
                detector placed at self._vertical_pixel_boresight, self._horizontal_pixel_boresight
        """
        if include_padding:
            return self._delta_omega_view

        return self._delta_omega_view[self._padding:self._padding + self._vert_pixel_width_no_padding,
                                      self._padding:self._padding + self._horz_pixel_width_no_padding]

    def set_disk(self,
                 disk_pixel_radius: float,
                 disk_horizontal_center: float,
                 disk_vertical_center: float,
                 distance_to_body_km: float):
        """
            Sets the current disk parameters and computes all parameters related to the position of the
            current disk.

            Parameters
            ----------
            disk_pixel_radius : float


            disk_horizontal_center : float
                Horizontal pixel coordinate of the image disk.

            disk_vertical_center : float
                Vertical pixel coordinate of the image disk.

            distance_to_body_km:
                Distance from the satellite to the sun/moon in km.
        """
        self._disk_pixel_radius = disk_pixel_radius
        self._hc = disk_horizontal_center + self._padding
        self._vc = disk_vertical_center + self._padding
        self._distance_to_body_km = distance_to_body_km
        DH, DV = self._compute_delta_pixels_relative_disk_center()
        self._horizontal_pixel_mesh_relative_to_disk_center = DH
        self._vertical_pixel_mesh_relative_to_disk_center = DV
        self._mu = self._compute_mu()
        self._mask = np.sqrt(1 - self._mu ** 2) <= self._disk_pixel_radius
        self._delta_omega_center, self._theta_center, self._phi_center = self.compute_delta_solid_angles(
            self._horizontal_pixel_mesh_relative_to_disk_center,
            self._vertical_pixel_mesh_relative_to_disk_center)

    def delta_omega_center(self):
        """
            Returns
            -------
            solid_angles : np.ndarray
                Detector-sized array of the solid angles of each pixel if the boresight were
                centered on the imaged disk.
        """
        return self._delta_omega_center

    def disk_mask(self, include_padding=False):
        """
            Parameters
            ----------
            include_padding: bool, Optional
                If true, return the array of the detector with padded pixels on the sides.

            Returns
            -------
            mask : np.ndarray
                True/False mask of detector pixels showing where the disk is on the detector.
        """
        if include_padding:
            return self._mask

        # crop the mask array to the correct size
        return self._mask[self._padding:self._padding + self._vert_pixel_width_no_padding,
                          self._padding:self._padding + self._horz_pixel_width_no_padding]

    def mu(self, include_padding=False):
        """
            Returns the radial parameter :math:`\mu` for each pixel on the detector.  The parameter is given by:

            .. math::
                \mu = \sqrt{1 - (\frac{r}{R})^2}

            where r is the observed 2D distance of the pixel from the center disk pixel, and R is the
            radius in pixels of the disk.

            Parameters
            ----------
            include_padding: bool, Optional
                 If true, return the array of the detector with padded pixels on the sides.

            Returns
            -------
            mu : np.ndarray
                Detector-sized array with values of mu for each pixel.  Pixels greater than the disk radius
                get a value of np.nan
        """

        if include_padding:
            return self._mu

        # crop the mu array to the correct size
        return self._mu[self._padding:self._padding + self._vert_pixel_width_no_padding,
                        self._padding:self._padding + self._horz_pixel_width_no_padding]

    def _compute_mu(self):
        R = np.sqrt(self._horizontal_pixel_mesh_relative_to_disk_center ** 2 +
                    self._vertical_pixel_mesh_relative_to_disk_center ** 2) / self._disk_pixel_radius
        R[R > 1.0] = np.nan
        mu = np.sqrt(1 - R ** 2)

        return mu

    def _get_delta_pixels_relative_boresight(self):
        dh = np.arange(0, self._horizontal_pixel_width) - self._horizontal_pixel_boresight
        dv = np.arange(0, self._vertical_pixel_width) - self._vertical_pixel_boresight

        DH, DV = np.meshgrid(dh, dv)
        return DH, DV

    def _compute_delta_pixels_relative_disk_center(self):
        """
            Gets the horizontal and vertical pixel indices relative to the centroid pixel.
        """
        dh = np.arange(0, self._horizontal_pixel_width) - self._hc
        dv = np.arange(0, self._vertical_pixel_width) - self._vc
        self._horizontal_pixel_rel_to_disk_center_1d = dh
        self._vertical_pixel_rel_to_disk_center_1d = dv

        DH, DV = np.meshgrid(dh, dv)
        return DH, DV

    # TODO this might be only computing a quarter of the solid angle of each pixel, need to fix
    def compute_delta_solid_angles(self,
                                   horizontal_pixel_mesh_rel_to_boresight,
                                   vertical_pixel_mesh_rel_to_boresight):
        """
            For a good visualization of the problem see:
            https://math.stackexchange.com/questions/2346018/how-do-i-compute-the-solid-angle-of-a-square-in-space-in-spherical-coordinates

            Computes the differential solid angles of each pixel of an image.

            Here the convention
            is that the boresight is the +x axis, the z axis points up on the detector, and the +y axis
            points to the LEFT to complete the right-handed coordinate system.

            The spherical angles :math:`\Theta, \phi` of each pixel and its top-left corner are computed,
            and the differential angles :math:`d\Theta, d\phi` are approximated by the difference between the
            corner and center angles of each pixel.  Since the typical HMI solar image is roughly 4096x4096 pixels
            covering a 0.55 degree span, this is a reasonable approximation.  The differential solid angles are
            computed by the formula:

            .. math::
                d\Omega = cos \Theta \cdot d\Theta \cdot d\phi

            Returns
            -------
            delta_omega : np.ndarray
               2D array of the differential solid angles corresponding to each pixel of the HMI
               image, same size as the HMI image.

            phi: np.ndarray
                2D array of azimuthal angle in spherical coordinates of each pixel

            theta: np.ndarray
                2D array of elevation in spherical coordinates of each pixel
        """
        y_center = horizontal_pixel_mesh_rel_to_boresight
        z_center = vertical_pixel_mesh_rel_to_boresight
        y_edge = y_center - 0.5
        z_edge = z_center + 0.5

        # xi = distance_to_body_km * self._body_radius_km / self._disk_pixel_radius  # distance in "pixels" to the sun
        xi = self._horizontal_pixel_width / 2 / np.arctan(np.deg2rad(self._horz_fov_deg))
        x = xi * np.ones_like(y_edge)

        theta = np.arctan(z_center / np.sqrt(x ** 2 + y_center ** 2))
        theta_edge = np.arctan(z_edge / np.sqrt(x ** 2 + y_edge ** 2))
        phi = np.arctan(y_center / x)
        phi_edge = np.arctan(y_edge / x)

        delta_theta = np.abs(theta_edge[1:, :] - theta_edge[:-1, :])
        delta_theta = np.append(delta_theta, delta_theta[-1, :][np.newaxis, ...], axis=0)

        delta_phi = np.abs(phi_edge[:, 1:] - phi_edge[:, :-1])
        delta_phi = np.append(delta_phi, delta_phi[:, -1][..., np.newaxis], axis=1)
        delta_omega = np.cos(theta) * delta_theta * delta_phi
        return delta_omega, phi, theta

    def compute_irradiance_for_viewing_angle(self, image: np.ndarray):
        """
            Corrects the total irradiance in arbitrary units for the solid angle of the actual viewing
            direction relative to the solid angle observed by an instrument with the boresight
            centered on the disk.  Assumes the pixel values are
            already proportional to irradiance (not radiance)
        """
        # slight scaling of the ideal center-viewed radiance:
        solid_angle_integrated_irradiance = np.nansum(image * self._delta_omega_view / self._delta_omega_center)
        solar_disk_solid_angle = np.nansum(self._delta_omega_view[self._mask])
        return solid_angle_integrated_irradiance, solar_disk_solid_angle, self._delta_omega_view

    def compute_radiance_per_solid_angle_steradian(self, image: np.ndarray):
        """
        TODO fill in, or delete, since not used anywhere
        """
        mask = self._mask[self._padding:self._padding + self._vertical_pixel_width,
                          self._padding:self._padding + self._horizontal_pixel_width]
        d_omega_v = self._delta_omega_view[self._padding:self._padding + self._vertical_pixel_width,
                                           self._padding:self._padding + self._horizontal_pixel_width]
        radiance_per_steradian = np.nansum(image[mask] / d_omega_v[mask])
        # returns the full disk size, even if off edge of detector:
        solar_disk_solid_angle = np.nansum(self._delta_omega_view[self._mask])
        return radiance_per_steradian, solar_disk_solid_angle


class SolarGeometry(DiskImageGeometry):
    """Implementation of :code:`DiskImageGeometry` for the sun"""

    def __init__(self, horizontal_pixel_width: int, vertical_pixel_width: int,
                 vertical_pixel_boresight: float, horizontal_pixel_boresight: float,
                 vert_fov_degrees: float, horz_fov_degrees: float,
                 padding: int = 0):
        super().__init__(horizontal_pixel_width, vertical_pixel_width,
                         horz_fov_degrees, vert_fov_degrees, SOLAR_RADIUS,
                         vertical_pixel_boresight, horizontal_pixel_boresight, padding)


class LunarGeometry(DiskImageGeometry):
    """Implementation of :code:`DiskImageGeometry` for the moon"""

    def __init__(self, horizontal_pixel_width: int, vertical_pixel_width: int,
                 vertical_pixel_boresight: float, horizontal_pixel_boresight: float,
                 vert_fov_degrees: float, horz_fov_degrees: float,
                 padding: int = 0):
        super().__init__(horizontal_pixel_width, vertical_pixel_width,
                         horz_fov_degrees, vert_fov_degrees, LUNAR_RADIUS,
                         vertical_pixel_boresight, horizontal_pixel_boresight, padding)


def compute_raster_box(horizontal_width: int,
                       vertical_width: int,
                       horizontal_pixels_per_raster_box: float,
                       vertical_pixels_per_raster_box: float,
                       padding_raster_boxes,
                       col_idx: int,
                       row_idx: int):
    """
    Parameters
    ----------
    horizontal_width : int

    vertical_width : int

    horizontal_pixels_per_raster_box : float

    vertical_pixels_per_raster_box : float

    padding_raster_boxes : int

    col_idx : int

    row_idx : int

    Returns
    -------
    raster_bounding_box : tuple(int, int, int, int)
        Tuple containing the start horizontal index, start vertical index, horizontal pixel
        span, and vertical pixel span, respectively.
    """

    rh = col_idx * horizontal_pixels_per_raster_box
    rv = row_idx * vertical_pixels_per_raster_box
    grid_boundaries_horz = np.arange(padding_raster_boxes * horizontal_pixels_per_raster_box,
                                     horizontal_width - padding_raster_boxes * horizontal_pixels_per_raster_box)
    grid_boundaries_vert = np.arange(padding_raster_boxes * vertical_pixels_per_raster_box,
                                     vertical_width - padding_raster_boxes * vertical_pixels_per_raster_box)
    raster_h_idx = np.abs(grid_boundaries_horz - rh).argmin()
    raster_v_idx = np.abs(grid_boundaries_vert - rv).argmin()
    raster_bounding_box = (raster_h_idx, raster_v_idx,
                           horizontal_pixels_per_raster_box,
                           vertical_pixels_per_raster_box)

    h_0 = raster_bounding_box[0]
    v_0 = raster_bounding_box[1]
    h_w = int(np.ceil(raster_bounding_box[2]))
    v_w = int(np.ceil(raster_bounding_box[3]))

    if h_0 + h_w > horizontal_width:
        h_w = horizontal_width - h_0
    if v_0 + v_w > vertical_width:
        v_w = vertical_width - v_0

    return v_0, h_0, v_w, h_w


# TODO implement least-squares center-finding algorithm

# TODO make this use as many points along the edge as manageable (not just 8)
def center_solar_disk(data, **kwargs):
    """
    Assumes a bright solar disk on a dark background (though not necessarily
    zero), and zeros in on the solar disk and its approximate center.  Uses
    matrix formulation for three points on an arc to find center:

    https://math.stackexchange.com/questions/213658/get-the-equation-of-a-circle-when-given-3-points

    Parameters
    ----------
    data : np.ndarray
        2D array of values representing an image of the solar disk, not necessarily
        centered on the image, or even have the full disk on the image

    detector_angular_width_degrees : Tuple of floats, Optional
        If passed in, makes use of the known angular size of the sun and determines the approximate
        radius of the sun in the image in terms of pixels.  The tuple is 2 elements: angular widths in degrees
        corresponding to the pixel dimensions of "data".

    distance_to_sun_km: float, Optional
        If not passed in, assumed to be average earth-to-sun distance.  Otherwise, is used to compute the approximate
        solar angular radius based on the physical solar radius of 696340 km

    Returns
    -------
    r : float
        Computed radius of the disk in pixels

    px : float
        x (horizontal) index to sub-pixel value of the center of the solar disk on the image

    py : float
        y (vertical) index to sub-pixel value of the center of the solar disk on the image

    algorithm_points : dict
        Points used for center-finding
    """

    # Remove any sunspots from the data, as best we can:
    data_with_zeros = data[np.isnan(data)] = 0.0
    edges = sobel(data)

    # find the approximate span of where the solar disk lies by
    # using max(edges)/3 as threshold
    approx_disk_indices = np.argwhere(edges > np.max(edges) / 3)

    xmin = np.min(approx_disk_indices[:, 1])
    xmax = np.max(approx_disk_indices[:, 1])
    ymin = np.min(approx_disk_indices[:, 0])
    ymax = np.max(approx_disk_indices[:, 0])

    approx_center_x = (xmax - xmin) / 2 + xmin
    approx_center_y = (ymax - ymin) / 2 + ymin
    approximate_radius = (np.max([np.abs(xmin - approx_center_x), np.abs(xmax - approx_center_x)]) +
                          np.max([np.abs(ymin - approx_center_y), np.abs(ymax - approx_center_y)])) / 2

    vw, hw = np.shape(data)
    if xmin == 0 or xmax == hw - 1:
        print('Warning: sun is past left/right edge of detector')
    if ymin == 0 or ymax == vw - 1:
        print('Warning: sun is past top/bottom edge of detector')

    override_radius = False

    if kwargs:
        # Override the radius calculation:
        override_radius = True
        detector_width_deg = kwargs['detector_angular_width_degrees']
        r_sun_km = 696340
        if 'distance_to_sun_km' in kwargs:
            d = kwargs['distance_to_sun_km']
        else:
            d = 147.24e6  # km
        solar_angular_radius = np.rad2deg(np.arctan(r_sun_km / d))
        pixels_per_deg = ((vw / detector_width_deg[0]) + (hw / detector_width_deg[1])) / 2
        approximate_radius = pixels_per_deg * solar_angular_radius
        approx_center_x = xmin + approximate_radius
        if xmin < 3:
            approx_center_x = xmax - approximate_radius
        approx_center_y = ymin + approximate_radius
        if ymin < 3:
            approx_center_y = ymax - approximate_radius
        if xmax - xmin < approximate_radius or ymax - ymin < approximate_radius:
            print('Warning: solar disk center is outside of detector')

    X, Y = np.meshgrid(range(len(edges)), range(len(edges[0])))
    R = np.sqrt((X - approx_center_x) ** 2 + (Y - approx_center_y) ** 2)
    edges_filtered = np.copy(edges)
    edges_filtered[np.abs(R - approximate_radius) / approximate_radius > 0.15] = 0.0

    y_width = ymax - ymin
    ag = {'horizontal_edge_points': [],
          'vertical_edge_points': [],
          'found_radii': [],
          'horizontal_center': [],
          'vertical_center': []}
    span = int(y_width / 3) - 1
    for i in range(2, span):
        frac = [i, span + i, 2 * span + i] / y_width
        rh, pxh, pyh, xeh, yeh = find_center_with_horizontal_cuts(edges_filtered, frac, xmax, xmin, ymax, ymin)
        ag['found_radii'].append(rh)
        ag['horizontal_center'].append(pxh)
        ag['vertical_center'].append(pyh)
        ag['horizontal_edge_points'].append(xeh)
        ag['vertical_edge_points'].append(yeh)

    x_width = xmax - xmin
    span = int(x_width / 3) - 1
    for i in range(2, span):
        frac = [i, span + i, 2 * span + i] / x_width
        rv, pxv, pyv, xev, yev = find_center_with_vertical_cuts(edges_filtered, frac, xmax, xmin, ymax, ymin)
        ag['found_radii'].append(rv)
        ag['horizontal_center'].append(pxv)
        ag['vertical_center'].append(pyv)
        ag['horizontal_edge_points'].append(xev)
        ag['vertical_edge_points'].append(yev)

    # Throw away anomalous points from sunspots near the edge of the disk
    valid_idx = np.argwhere(np.abs(ag['found_radii'] - np.mean(ag['found_radii'])) < np.std(ag['found_radii']))

    valid_idx = valid_idx.flatten()
    ag['found_radii'] = np.asarray(ag['found_radii'])[valid_idx]
    ag['horizontal_center'] = np.asarray(ag['horizontal_center'])[valid_idx]
    ag['vertical_center'] = np.asarray(ag['vertical_center'])[valid_idx]
    ag['horizontal_edge_points'] = np.asarray(ag['horizontal_edge_points'])[valid_idx]
    ag['vertical_edge_points'] = np.asarray(ag['vertical_edge_points'])[valid_idx]

    # Repeat once the biggest outliers are gone:
    valid_idx = np.argwhere(np.abs(ag['found_radii'] - np.mean(ag['found_radii'])) < np.std(ag['found_radii']))

    valid_idx = valid_idx.flatten()
    ag['found_radii'] = ag['found_radii'][valid_idx]
    ag['horizontal_center'] = ag['horizontal_center'][valid_idx]
    ag['vertical_center'] = ag['vertical_center'][valid_idx]
    ag['horizontal_edge_points'] = ag['horizontal_edge_points'][valid_idx]
    ag['vertical_edge_points'] = ag['vertical_edge_points'][valid_idx]

    # Find the average of the centers and radii, with half-pixel correction
    r = np.mean(ag['found_radii'])
    px = np.mean(ag['horizontal_center'])
    py = np.mean(ag['vertical_center'])

    if override_radius:
        r = approximate_radius

    return r, px, py, ag


def find_center_with_horizontal_cuts(edges, fractions, xmax, xmin, ymax, ymin):
    y_width = ymax - ymin
    y1 = int(ymin + fractions[0] * y_width)
    y2 = int(ymin + fractions[1] * y_width)
    y3 = int(ymin + fractions[2] * y_width)

    x1 = np.nanargmax(edges[y1, xmin:xmax]) + xmin
    x2 = np.nanargmax(edges[y2, xmin:xmax]) + xmin
    x3 = np.nanargmax(edges[y3, xmin:xmax]) + xmin

    r1sq = x1 ** 2 + y1 ** 2
    r2sq = x2 ** 2 + y2 ** 2
    r3sq = x3 ** 2 + y3 ** 2
    A = np.array([[0, 0, 0, 1],
                  [r1sq, x1, np.floor(y1), 1],
                  [r2sq, x2, np.floor(y2), 1],
                  [r3sq, x3, np.floor(y3), 1]])
    m11 = getMatrixMinor(A, 1, 1)
    m12 = getMatrixMinor(A, 1, 2)
    m13 = getMatrixMinor(A, 1, 3)
    m14 = getMatrixMinor(A, 1, 4)
    px = _correct_pixel(0.5 * m12 / m11)
    py = _correct_pixel(-0.5 * m13 / m11)
    r = np.sqrt(px ** 2 + py ** 2 + m14 / m11)
    return r, px, py, (x1, x2, x3), (y1, y2, y3)


def _correct_pixel(pixel_val):
    return pixel_val  # + 0.5


def find_center_with_vertical_cuts(edges, fractions, xmax, xmin, ymax, ymin):
    x_width = xmax - xmin
    x1 = int(xmin + fractions[0] * x_width)
    x2 = int(xmin + fractions[1] * x_width)
    x3 = int(xmin + fractions[2] * x_width)

    y1 = np.nanargmax(edges[ymin:ymax, x1]) + ymin
    y2 = np.nanargmax(edges[ymin:ymax, x2]) + ymin
    y3 = np.nanargmax(edges[ymin:ymax, x3]) + ymin
    r1sq = x1 ** 2 + y1 ** 2
    r2sq = x2 ** 2 + y2 ** 2
    r3sq = x3 ** 2 + y3 ** 2
    A = np.array([[0, 0, 0, 1],
                  [r1sq, x1, y1, 1],
                  [r2sq, x2, y2, 1],
                  [r3sq, x3, y3, 1]])
    m11 = getMatrixMinor(A, 1, 1)
    m12 = getMatrixMinor(A, 1, 2)
    m13 = getMatrixMinor(A, 1, 3)
    m14 = getMatrixMinor(A, 1, 4)
    px = _correct_pixel(0.5 * m12 / m11)
    py = _correct_pixel(-0.5 * m13 / m11)
    r = np.sqrt(px ** 2 + py ** 2 + m14 / m11)
    return r, px, py, (x1, x2, x3), (y1, y2, y3)


def getMatrixMinor(m, i, j):
    # ones indexed
    row = np.arange(1, 5)
    col = np.arange(1, 5)

    minor = np.array([r[col != j] for r in m[row != i]])
    return np.linalg.det(minor)
