import os

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.patches import Circle
from sasktran import SolarSpectrum
from scipy.optimize import least_squares

from altius_l1pp_inflight.configuration.altius_config import AltiusL1ppInflightConfig
from altius_l1pp_inflight.scene.altius_scene import AltiusSolarScene
from altius_l1pp_inflight.scene.disk_geometry import D_1AU, SolarGeometry
from altius_l1pp_inflight.scene.limbdark_mu_pierceandslaughter1977 import LimbDarkening_PS1977
from altius_l1pp_inflight.srf.SrfInterface import SrfInterface


class SolarCalibrationUtil:
    """
        Utility class that handles common calculations in the PRNU calibration, like flattening
        the image due to limb darkening, and computing the irradiance of the disk.  Handles the
        geometry portion using the :code:`SolarGeometry` class.
    """

    def __init__(self, solar_geometry: SolarGeometry, srf: SrfInterface, padding_multiplier=0):
        """
            Parameters
            ----------
            solar_geometry: SolarGeometry
                Instance of SolarGeometry for the detector image dimensions.  The instance used
                by this class adds padding onto the sides of the image so images of the sun that
                are partially off the detector still have their irradiance computed correctly.

            srf: SrfInterface
                Instance of an SrfInterface object, specifies the spectral response function in the
                band of interest for this calibration.
        """
        self._solar_geometry = solar_geometry.with_padding(padding_multiplier)
        self._srf = srf

    def flatten_solar_disk_image(self, image, horizontal_center, vertical_center,
                                 radius, satellite_to_sun_distance_km, wavelength,
                                 bounds=None):
        """
            Takes an image of the sun, and given the found center and radius,
            Parameters
            ----------
            image: np.ndarray[nv, nh]
                The image with the solar disk to have limb darkening removed.  It does not need to be an
                entire ALTIUS detector field - it is more reasonable to pass in the region of the sun on the image
                which needs to be flattened (the "raster box")

            horizontal_center: float
                Position of the solar center on the image in the horizontal dimension, to sub-pixel accuracy.

            vertical_center: float
                Position of the solar center on the image in the vertical dimension, to sub-pixel accuracy.

            radius: float
                Radius of the solar disk for the image.  Since it is potentially a cropped image of the interior
                of the sun, the radius is allowed to be larger than the width of the image.

            satellite_to_sun_distance_km : float
                Distance from the satellite to the sun in km.

            wavelength: float
                Wavelength in nm

            bounds: Tuple, Optional
                If passed in, is in format (v0, h0, vw, hw), where the first and second parameters are the vertical
                and horizontal starting pixels, and the third and fourth parameters are vertical and horizontal
                spans.  This is used for cropping the final output flattened image and its estimated error.

            Returns
            -------
            dictionary, which contains the two following fields:

            flattened_image: np.ndarray
                The image with limb darkening removed, based off the center and radius (and wavelength)
                passed in.  Reduced to the width specified in bounds if present.

            error_estimate: np.ndarray
                The estimate of the error on the image by taking the difference of a least-squares fitted
                limb darkening profile and the actual input image.  Reduced to the width specified in bounds
                if present.
        """
        mu_profile = np.linspace(0, 1.0, 50)
        limb_darkening = LimbDarkening_PS1977()
        ld_wl_profile = limb_darkening.limb_darkening(wavelength, mu_profile)

        self._solar_geometry.set_disk(radius, horizontal_center, vertical_center, satellite_to_sun_distance_km)
        mu = self._solar_geometry.mu()

        ld = np.interp(mu, mu_profile, ld_wl_profile)

        flattened_image = image / ld

        if bounds is not None:
            flattened_image = flattened_image[bounds[0]:bounds[0] + bounds[2], bounds[1]:bounds[1] + bounds[3]]
            cropped_image = image[bounds[0]:bounds[0] + bounds[2], bounds[1]:bounds[1] + bounds[3]]
            cropped_ld = ld[bounds[0]:bounds[0] + bounds[2], bounds[1]:bounds[1] + bounds[3]]
        else:
            cropped_image = image
            cropped_ld = ld

        # estimate the error by fitting the limb darkening to the image.
        # This only works decently for the rectangular region inside the sun, avoiding the edges
        result = least_squares(fun=self._compute_residual, x0=np.array([np.nanmean(flattened_image)]),
                               args=(cropped_ld, cropped_image), loss='cauchy')

        error_estimate = result.x * cropped_ld - cropped_image

        return {'flattened_image': flattened_image, 'error_estimate': error_estimate}

    def _compute_residual(self, scale_factor, limb_darkening, image):
        residual = scale_factor * limb_darkening.flatten() - image.flatten()
        return residual

    def compute_solar_irradiance(self,
                                 wavelength: float,
                                 solar_distance: float):
        """
            Parameters
            ----------
            wavelength: float
                The calibration wavelength in nm.

            solar_distance: float
                The distance from the satellite to the sun in km.

            Returns
            -------
            with_srf: np.ndarray
                The high-resolution solar spectrum convolved with the SRF at the calibration wavelength,
                so is a 1-element array.  Can technically be any units, but should be in photons/s/cm2/nm/sr
        """
        solar = SolarSpectrum()
        wl = self._srf.get_wavelength_grid(0.005)
        spectrum = solar.irradiance(wl, solardistance=solar_distance / D_1AU)
        with_srf = self._srf.convolve_spectrum_with_srf(np.array([wavelength]), wl, spectrum)
        return with_srf

    def model_noiseless_solar_disk(self, cv, ch, r, wavelength, solar_distance, with_padding):
        self._solar_geometry.set_disk(r, ch, cv, solar_distance)

        mu_profile = np.linspace(0, 1.0, 50)
        limb_darkening = LimbDarkening_PS1977()
        ld_wl_profile = limb_darkening.limb_darkening(wavelength, mu_profile)

        mu = self._solar_geometry.mu()
        ld = np.interp(mu, mu_profile, ld_wl_profile)

        # delta_omega = self._solar_geometry.delta_omega_view(True)
        # irradiance = self.compute_solar_irradiance(wavelength, solar_distance)
        # scale = irradiance / np.nansum(ld * delta_omega)
        scale = self.flat_solar_disk_model(cv, ch, r, wavelength, solar_distance)
        # vl, vu, hl, hu = self._solar_geometry.get_image_bounds()

        return ld * scale

    def flat_solar_disk_model(self, cv, ch, r,
                              wavelength, solar_distance):
        """
            This function serves as the model for the ideal sun for relative calibration.
            It re-creates the sun at the detector location of the actual imaged sun, using the
            Pierce & Slaughter limb darkening, and computes the solid angles of each pixel.
            The input irradiance (total solar disk irradiance) is divided by the integral
            of the limb darkening and solid angles to yield the scale factor to normalize the
            PRNU map.

            Parameters
            ----------

            cv: float
                Center of the sun on the detector in the vertical, to sub-pixel accuracy.
                Half-pixels are the centers of pixels

            ch: float
                Center of the sun on the detector in the horizontal, to sub-pixel accuracy.

            r: float
                Radius of the solar disk in pixels, to sub-pixel accuracy.

            wavelength: float
                Calibration wavelength in nm.

            solar_distance: float
                Satellite-to-sun distance in km.

            Returns
            -------
            scale: float
                The value to scale the limb darkening on a pixel-per-steradian basis to give
                the correct irradiance.  This number is used to scale down the PRNU from absolute values.
        """
        self._solar_geometry.set_disk(r, ch, cv, solar_distance)

        mu_profile = np.linspace(0, 1.0, 50)
        limb_darkening = LimbDarkening_PS1977()
        ld_wl_profile = limb_darkening.limb_darkening(wavelength, mu_profile)

        mu = self._solar_geometry.mu(include_padding=True)
        ld = np.interp(mu, mu_profile, ld_wl_profile)

        delta_omega = self._solar_geometry.delta_omega_view(include_padding=True)
        irradiance = self.compute_solar_irradiance(wavelength, solar_distance)

        scale = irradiance / np.nansum(ld * delta_omega)

        return scale


def get_solar_prnu_simulated_observation_dir(config: AltiusL1ppInflightConfig, scenario: str = 'NO_SUNSPOTS'):
    num_vert_scans = config.get_num_vertical_raster_scans()
    num_horz_scans = config.get_num_horizontal_raster_scans()
    raster_grid_str = str(num_horz_scans) + 'x' + str(num_vert_scans)
    home_dir = config.get_processed_solar_prnu_file_directory()
    output_directory = os.path.join(home_dir, 'altius_observation', raster_grid_str, scenario)
    if not os.path.exists(output_directory):
        os.makedirs(output_directory)
    return output_directory


def get_solar_prnu_calibration_dir(config: AltiusL1ppInflightConfig, scenario: str = 'NO_SUNSPOTS'):
    num_vert_scans = config.get_num_vertical_raster_scans()
    num_horz_scans = config.get_num_horizontal_raster_scans()
    raster_grid_str = str(num_horz_scans) + 'x' + str(num_vert_scans)
    home_dir = config.get_processed_solar_prnu_file_directory()
    output_directory = os.path.join(home_dir, 'calibration', raster_grid_str, scenario)
    return output_directory


# Random utils (likely deprecated)

# Welford algorithm for mean and variance
# For a new value newValue, compute the new count, new mean, the new M2.
# mean accumulates the mean of the entire dataset
# M2 aggregates the squared distance from the mean
# count aggregates the number of samples seen so far
def update(existingAggregate, newValue):
    (count, mean, M2) = existingAggregate
    count += 1
    delta = newValue - mean
    mean += delta / count
    delta2 = newValue - mean
    M2 += delta * delta2
    return count, mean, M2


# Retrieve the mean, variance and sample variance from an aggregate
def finalize(existingAggregate):
    (count, mean, M2) = existingAggregate
    if count < 2:
        return float("nan")
    else:
        (mean, variance, sampleVariance) = (mean, M2 / count, M2 / (count - 1))
        return mean, variance, sampleVariance


def compare_disk_centers(altius_scene: AltiusSolarScene, found_horz_center, found_vert_center, found_radius,
                         horz_corrected_center, vert_corrected_center):
    fig, ax = plt.subplots()
    ax.imshow(altius_scene.get_altius_image())
    ax.plot(found_horz_center, found_vert_center, 'r.')
    ax.plot(altius_scene.get_expected_altius_solar_center_horz(),
            altius_scene.get_expected_altius_solar_center_vert(), 'bx')
    ax.plot(horz_corrected_center, vert_corrected_center, 'g*')

    plt.legend(['Initial center', 'Ideal Center', 'Entropy-corrected Center'])

    circ = Circle((found_horz_center, found_vert_center), found_radius, fill=False, color='r')
    circ_e = Circle((altius_scene.get_expected_altius_solar_center_horz(),
                     altius_scene.get_expected_altius_solar_center_vert()),
                    altius_scene.get_expected_altius_solar_radius(), fill=False, color='b')
    circ_cor = Circle((horz_corrected_center, vert_corrected_center), found_radius, fill=False, color='g')
    ax.add_patch(circ)
    ax.add_patch(circ_e)
    ax.add_patch(circ_cor)
    plt.title('Comparison of found center vs. ideal center vs. corrected center')

    plt.show()
