import logging
from datetime import datetime

import matplotlib.pyplot as plt
import numpy as np
import sktime.mjd as mjd_conv
from scipy.interpolate import RegularGridInterpolator

from altius_l1pp_inflight.scene.disk_geometry import SolarGeometry
from altius_l1pp_inflight.scene.generation.solar.hmi_files import SolarDiskRadianceResult
from altius_l1pp_inflight.scene.limbdark_mu_pierceandslaughter1977 import LimbDarkening_PS1977
from altius_l1pp_inflight.scene.sunspot_detection import detect_sunspots


class SimulatedSolarDiskInterpolator:
    """
    Reads in real HMI data and synthesizes ideal solar data using a simple limb darkening model

    Parameters
    ----------
    all_hmi_results : np.ndarray
       Array of type SolarDiskRadianceResult for the actual HMI data

    Returns
    -------
    R : np.ndarray
       <description>
    """

    def __init__(self, all_hmi_results: np.ndarray):

        print('Characterizing noise in granulation and time for synthesis of data')
        self._ideal_solar_images = self.simulate_solar_disks(all_hmi_results)

        print('Creating solar disk interpolator for scenario')
        self._all_hmi_results = all_hmi_results

        print('Establishing pixel interpolation grid using first HMI image')
        first_sim: SolarDiskRadianceResult = self._ideal_solar_images[0]

        # can skip the step of interpolating to same grid, since it's ideal data
        # chosen regular grid to interpolate to in x-y pixels
        hmi_center_pixel = first_sim.center_pixel()
        h_size, v_size = first_sim.radiance().shape

        self._solar_geometry = SolarGeometry(h_size, v_size, hmi_center_pixel[1], hmi_center_pixel[0], 0.5,
                                             0.5)  # FOV doesn't matter here
        self._solar_geometry.set_disk(first_sim.average_solar_radius_in_pixels(), hmi_center_pixel[0],
                                      hmi_center_pixel[1], first_sim.actual_solar_distance())
        self._DH, self._DV = self._solar_geometry.get_delta_pixels_relative_disk_center()
        dh = self._DH[0, :].flatten()
        dv = self._DV[:, 0].flatten()
        gridded_radiances = []
        gridded_radiances.append(first_sim.radiance())
        self._mjd_times = []
        self._mjd_times.append(mjd_conv.datetime_to_mjd(first_sim.time()))
        utc_times = [first_sim.time()]

        for idx, hmi in enumerate(all_hmi_results[1:]):
            if not isinstance(hmi, SolarDiskRadianceResult):
                raise AttributeError('SolarDiskInterpolator requires an array of HmiSolarDiskResult')
            # h_size, v_size = hmi.radiance().shape
            # x_cc, y_cc = hmi.center_pixel()
            # dx_p, dy_p = get_delta_pixels_relative_centroid(x_cc, y_cc, h_size, v_size)
            # logging.info('Interpolating HMI image ', idx, ' to first image grid')
            # f_spatial = interp2d(dx_p, dy_p, hmi.radiance())
            # gridded_radiances.append(f_spatial(dh, dv))
            self._mjd_times.append(mjd_conv.datetime_to_mjd(hmi.time()))
            utc_times.append(hmi.time())

        self._gridded_radiances = [r.radiance_disk for r in self._ideal_solar_images]
        self._mjd_times = np.array(self._mjd_times)
        # Now they're all on the same [dx, dy] grid, set up the interpolation functions in time
        print('Setting up interpolator between images at times: ', utc_times)
        self._time_interp_function = RegularGridInterpolator(points=(np.array(self._mjd_times), dv, dh),
                                                             values=np.array(self._gridded_radiances))

        # TODO other parameters of the common grid, like center position, etc.

    def get_simulated_ideal_grid_images(self) -> np.ndarray:
        return self._ideal_solar_images

    def grid_mjd_times(self) -> np.ndarray:
        return self._mjd_times

    def solar_disk_average_radius_in_pixels(self):
        return self._all_hmi_results[0].average_solar_radius_in_pixels()  # these should all be the same TODO enforce

    def solar_radius_apparent_angle(self):
        return self._all_hmi_results[0].angular_radius_degrees()

    def solar_disk_angles(self):
        return self._all_hmi_results[0].theta(), self._all_hmi_results[0].phi()

    def solar_disk_center_pixels(self):
        # TODO verify these are the same (should be since they're interpolated on to the same grid)
        return self._ideal_solar_images[0].center_pixel()

    def interpolate_disk(self, time: datetime) -> SolarDiskRadianceResult:
        """
        Uses the interpolators set up when initializing the class to interpolate the disk radiance to
        the time requested

        Parameters
        ----------
        time : datetime
           Time of the measurement to interpolate to.

        Returns
        -------
        SolarDiskRadianceResult
           The solar disk interpolated at every pixel to the time requested.
        """
        mjd_interp_time = mjd_conv.datetime_to_mjd(time)
        first_sim: SolarDiskRadianceResult = self._ideal_solar_images[0]
        x_c, y_c = first_sim.center_pixel()

        logging.info('Interpolating disk radiance to ', time)
        interp_radiance = self._time_interp_function((mjd_interp_time, self._DV, self._DH), method='linear')

        # TODO interpolate all the other results, like actual solar distance, etc.
        average_radius_pixels = first_sim.average_solar_radius_in_pixels()
        actual_solar_distance = self.actual_solar_distance(time)
        mu = first_sim.mu_solar_disk()
        wavelength = first_sim.wavelength()
        solar_radius_apparent_angle = first_sim.angular_radius_degrees()

        return SolarDiskRadianceResult(interp_radiance, vertical_disk_pixel_centroid=y_c,
                                       horizontal_disk_pixel_centroid=x_c, average_radius_pixels=average_radius_pixels,
                                       actual_solar_distance_km=actual_solar_distance, mu=mu, time=time,
                                       wavelength=wavelength,
                                       solar_radius_apparent_angle_deg=solar_radius_apparent_angle)

    def actual_solar_distance(self, time: datetime) -> float:

        interp_array = np.interp(x=mjd_conv.datetime_to_mjd(time), xp=self._mjd_times,
                                 fp=[hmi.actual_solar_distance() for hmi in self._all_hmi_results])
        return interp_array

    @staticmethod
    def simulate_solar_disks(all_hmi_results: np.ndarray, show_figures=False):

        mu_profile = np.linspace(0, 1, 50)
        ld_model = LimbDarkening_PS1977()

        all_flattened = []
        first_hmi = all_hmi_results[0]

        # Choose the box:
        solar_radius_pixels = first_hmi.average_solar_radius_in_pixels()
        hc, vc = first_hmi.center_pixel()
        lh = int(hc - solar_radius_pixels / 2)
        uh = int(hc + solar_radius_pixels / 2)
        lv = int(vc - solar_radius_pixels / 2)
        uv = int(vc + solar_radius_pixels / 2)

        for hmi in all_hmi_results:
            # apply limb darkening removal
            ld_profile = ld_model.limb_darkening(hmi.wavelength(), mu_profile)
            ld = np.interp(hmi.mu_solar_disk(), mu_profile, ld_profile)
            flattened = hmi.radiance() / ld
            flattened = flattened[lv:uv, lh:uh]
            _, _, _, flattened = detect_sunspots(flattened, threshold_value=0.6)  # places nan's in sunspot locations
            all_flattened.append(flattened)

        all_flattened = np.array(all_flattened)

        # Total variation with time and space:
        mean = np.nanmean(all_flattened)
        std_dev = np.nanstd(all_flattened)

        sim_results = SimulatedSolarDiskInterpolator._create_simulated_disks(all_hmi_results,
                                                                             solar_radius_pixels, mean,
                                                                             std_dev)
        if show_figures:
            plt.figure()
            plt.imshow(np.nanmean(all_flattened, axis=0))
            plt.title('Mean of Central Region HMI Images')

            lhf = int(hc - solar_radius_pixels / 8)
            uhf = int(hc + solar_radius_pixels / 8)
            lvf = int(vc - solar_radius_pixels / 8)
            uvf = int(vc + solar_radius_pixels / 8)

            xlim = [mean - 3 * std_dev, mean + 3 * std_dev]
            all_flattened_sim = []
            for r, sim in zip(all_hmi_results, sim_results):
                fig, (ax1, ax2) = plt.subplots(1, 2)
                im = ax1.imshow(sim.radiance_disk)
                fig.colorbar(im)
                ax2.hist(sim.radiance_disk[lvf:uvf, lhf:uhf].flatten(), bins=200)
                ax2.set_xlim(xlim)
                plt.title('Central Region of Simulated Disk ' + sim.time().strftime("%m/%d/%Y, %H:%M:%S"))

                plt.figure()
                plt.plot(r.radiance_disk[2247, :], 'r--', alpha=0.5)
                plt.plot(sim.radiance_disk[2247, :], alpha=0.5)
                plt.legend(['HMI Data', 'Simulated'])
                plt.title('Comparing Solar Disk Cross Sections' + sim.time().strftime("%m/%d/%Y, %H:%M:%S"))

                ld_profile = ld_model.limb_darkening(sim.wavelength(), mu_profile)
                ld = np.interp(sim.mu_solar_disk(), mu_profile, ld_profile)
                flat_sim = sim.radiance() / ld
                flat_sim = flat_sim[lv:uv, lh:uh]
                all_flattened_sim.append(flat_sim)

            all_flattened_sim = np.array(all_flattened_sim)
            plt.figure()
            plt.hist(all_flattened.flatten(), bins=200, alpha=0.5, density=True)
            plt.hist(all_flattened_sim.flatten(), bins=200, alpha=0.5, density=True)
            plt.legend(['HMI Data', 'Simulated'])
            plt.xlim(xlim)
            plt.title('Distribution of Actual HMI Solar Central Image vs. Simulated')
            plt.show()

        return sim_results

    @staticmethod
    def _create_simulated_disks(all_hmi_results, solar_radius_pixels, mean, std_dev):
        hc = 2047.5
        vc = 2047.5
        hw = 4096
        vw = 4096

        solar_geometry = SolarGeometry(hw, vw, vc, hc, 0.5, 0.5)  # fov doesn't matter here

        simulated_disks = []
        ld_model = LimbDarkening_PS1977()
        mu_profile = np.linspace(0, 1, 50)
        for i, hmi in enumerate(all_hmi_results):
            solar_geometry.set_disk(hmi.average_solar_radius_in_pixels(), hc, vc,
                                    hmi.actual_solar_distance())
            mu = solar_geometry.mu()
            rad = hmi.radiance()
            ld_profile = ld_model.limb_darkening(hmi.wavelength(), mu_profile)
            ld = np.interp(hmi.mu_solar_disk(), mu_profile, ld_profile)
            flat_sim = rad / ld
            # find the sunspots, then add them- in later:
            _, _, _, sunspots_removed = detect_sunspots(flat_sim / mean, threshold_value=0.3)

            sunspots_only = np.empty_like(rad)
            sunspots_only[:] = np.nan
            sunspots_only[np.isnan(sunspots_removed)] = rad[np.isnan(sunspots_removed)] / mean  # only set the sunspots
            sunspots_only = sunspots_only[int(vc - vw / 4):int(vc + vw / 4), int(hc - hw / 4):int(hc + hw / 4)]

            ld_profile = ld_model.limb_darkening(hmi.wavelength(), mu_profile)
            sim = np.interp(mu, mu_profile, ld_profile)
            sim[~np.isnan(sim)] = np.random.normal(sim[~np.isnan(sim)], std_dev / mean)
            sim_box = sim[int(vc - vw / 4):int(vc + vw / 4), int(hc - hw / 4):int(hc + hw / 4)]
            sim_box[~np.isnan(sunspots_only)] = sunspots_only[~np.isnan(sunspots_only)]  # add in the sunspots
            sim[int(vc - vw / 4):int(vc + vw / 4), int(hc - hw / 4):int(hc + hw / 4)] = sim_box
            sim[np.isnan(sim)] = 0.0
            sim *= np.mean(mean)  # arbitrary, but puts it back similar to hmi files

            result = SolarDiskRadianceResult(radiance_disk=sim,
                                             vertical_disk_pixel_centroid=vc,
                                             horizontal_disk_pixel_centroid=hc,
                                             average_radius_pixels=solar_radius_pixels,
                                             actual_solar_distance_km=hmi.actual_solar_distance(),
                                             mu=mu, time=hmi.time(), wavelength=hmi.wavelength(),
                                             solar_radius_apparent_angle_deg=hmi.angular_radius_degrees())
            simulated_disks.append(result)
        return np.array(simulated_disks)
