import numpy as np
from skimage.filters import sobel


# im = image.imread('altius_l1pp_inflight/+scene/generation/lunar/fullmoon.png')
# im = 0.2989*im[:, :, 0] + 0.5870*im[:, :, 1] + 0.1140*im[:, :, 2]


def center_lunar_image(data, **kwargs):
    """
    TODO incorporate DSK from Didier (uses least squares of edge)
    Assumes a bright lunar disk on a dark background (though not necessarily
    zero), and zeros in on the solar disk and its approximate center.  Uses
    matrix formulation for three points on an arc to find center:

    https://math.stackexchange.com/questions/213658/get-the-equation-of-a-circle-when-given-3-points

    This is borrowed from the solar disk centering code and catered
    to the specific lunar image used for simulation

    Parameters
    ----------
    data : np.ndarray
        2D array of values representing the image of the moon

    Returns
    -------
    r : float
        Computed radius of the disk in pixels

    px : float
        x (horizontal) index to sub-pixel value of the center of the lunar disk on the image

    py : float
        y (vertical) index to sub-pixel value of the center of the lunar disk on the image
    """

    edges = sobel(data)

    # find the approximate span of where the lunar disk lies by
    # using max(edges)/3 as threshold
    approx_disk_indices = np.argwhere(edges > np.max(edges) / 3)

    xmin = np.min(approx_disk_indices[:, 1])
    xmax = np.max(approx_disk_indices[:, 1])
    ymin = np.min(approx_disk_indices[:, 0])
    ymax = np.max(approx_disk_indices[:, 0])

    approx_center_x = (xmax - xmin) / 2 + xmin
    approx_center_y = (ymax - ymin) / 2 + ymin
    approximate_radius = (np.max([np.abs(xmin - approx_center_x), np.abs(xmax - approx_center_x)]) +
                          np.max([np.abs(ymin - approx_center_y), np.abs(ymax - approx_center_y)])) / 2

    vw, hw = np.shape(data)
    if xmin == 0 or xmax == hw - 1:
        print('Warning: moon is past left/right edge of detector')
    if ymin == 0 or ymax == vw - 1:
        print('Warning: moon is past top/bottom edge of detector')

    override_radius = False

    if kwargs:
        # Override the radius calculation:
        override_radius = True
        detector_width_deg = kwargs['detector_angular_width_degrees']
        r_sun_km = 696340
        if 'distance_to_sun_km' in kwargs:
            d = kwargs['distance_to_sun_km']
        else:
            d = 147.24e6  # km
        solar_angular_radius = np.rad2deg(np.arctan(r_sun_km / d))
        pixels_per_deg = ((vw / detector_width_deg[0]) + (hw / detector_width_deg[1])) / 2
        approximate_radius = pixels_per_deg * solar_angular_radius
        approx_center_x = xmin + approximate_radius
        if xmin < 3:
            approx_center_x = xmax - approximate_radius
        approx_center_y = ymin + approximate_radius
        if ymin < 3:
            approx_center_y = ymax - approximate_radius
        if xmax - xmin < approximate_radius or ymax - ymin < approximate_radius:
            print('Warning: solar disk center is outside of detector')

    X, Y = np.meshgrid(range(len(edges)), range(len(edges[0])))
    R = np.sqrt((X - approx_center_x) ** 2 + (Y - approx_center_y) ** 2)
    edges_filtered = np.copy(edges)
    edges_filtered[np.abs(R - approximate_radius) / approximate_radius > 0.05] = 0.0

    fractions = [1 / 9, 4 / 9, 7 / 9]
    r1h, px1h, py1h, x1eh, y1eh = find_center_with_horizontal_cuts(edges_filtered, fractions, xmax, xmin, ymax, ymin)
    r1v, px1v, py1v, x1ev, y1ev = find_center_with_vertical_cuts(edges_filtered, fractions, xmax, xmin, ymax, ymin)
    fractions = [2 / 9, 5 / 9, 8 / 9]
    r2h, px2h, py2h, x2eh, y2eh = find_center_with_horizontal_cuts(edges_filtered, fractions, xmax, xmin, ymax, ymin)
    r2v, px2v, py2v, x2ev, y2ev = find_center_with_vertical_cuts(edges_filtered, fractions, xmax, xmin, ymax, ymin)
    fractions = [3 / 9, 4 / 9, 7 / 9]
    r3h, px3h, py3h, x3eh, y3eh = find_center_with_horizontal_cuts(edges_filtered, fractions, xmax, xmin, ymax, ymin)
    r3v, px3v, py3v, x3ev, y3ev = find_center_with_vertical_cuts(edges_filtered, fractions, xmax, xmin, ymax, ymin)
    fractions = [1 / 9, 6 / 9, 8 / 9]
    r4h, px4h, py4h, x4eh, y4eh = find_center_with_horizontal_cuts(edges_filtered, fractions, xmax, xmin, ymax, ymin)
    r4v, px4v, py4v, x4ev, y4ev = find_center_with_vertical_cuts(edges_filtered, fractions, xmax, xmin, ymax, ymin)

    # Find the average of the centers and radii, with half-pixel correction
    r = (r1h + r1v + r2h + r2v + r3h + r3v + r4h + r4v) / 8
    px = (px1h + px1v + px2h + px2v + px3h + px3v + px4h + px4v) / 8
    py = (py1h + py1v + py2h + py2v + py3h + py3v + py4h + py4v) / 8

    if override_radius:
        r = approximate_radius

    algorithm_points = {
        'horizontal_edge_points': [x1eh, x1ev, x2eh, x2ev, x3eh, x3ev, x4eh, x4ev],
        'vertical_edge_points': [y1eh, y1ev, y2eh, y2ev, y3eh, y3ev, y4eh, y4ev],
        'found_radii': [r1h, r1v, r2h, r2v, r3h, r3v, r4h, r4v],
        'horizontal_center': [px1h, px1v, px2h, px2v, px3h, px3v, px4h, px4v],
        'vertical_center': [py1h, py1v, py2h, py2v, py3h, py3v, py4h, py4v]}

    return r, px, py, algorithm_points


def find_center_with_horizontal_cuts(edges, fractions, xmax, xmin, ymax, ymin):
    y_width = ymax - ymin
    y1 = int(ymin + fractions[0] * y_width)
    y2 = int(ymin + fractions[1] * y_width)
    y3 = int(ymin + fractions[2] * y_width)

    x1 = np.nanargmax(edges[y1, xmin:xmax]) + xmin
    x2 = np.nanargmax(edges[y2, xmin:xmax]) + xmin
    x3 = np.nanargmax(edges[y3, xmin:xmax]) + xmin

    r1sq = x1 ** 2 + y1 ** 2
    r2sq = x2 ** 2 + y2 ** 2
    r3sq = x3 ** 2 + y3 ** 2
    A = np.array([[0, 0, 0, 1],
                  [r1sq, x1, np.floor(y1), 1],
                  [r2sq, x2, np.floor(y2), 1],
                  [r3sq, x3, np.floor(y3), 1]])
    m11 = getMatrixMinor(A, 1, 1)
    m12 = getMatrixMinor(A, 1, 2)
    m13 = getMatrixMinor(A, 1, 3)
    m14 = getMatrixMinor(A, 1, 4)
    px = _correct_pixel(0.5 * m12 / m11)
    py = _correct_pixel(-0.5 * m13 / m11)
    r = np.sqrt(px ** 2 + py ** 2 + m14 / m11)
    return r, px, py, (x1, x2, x3), (y1, y2, y3)


def _correct_pixel(pixel_val):
    return pixel_val  # + 0.5


def find_center_with_vertical_cuts(edges, fractions, xmax, xmin, ymax, ymin):
    x_width = xmax - xmin
    x1 = int(xmin + fractions[0] * x_width)
    x2 = int(xmin + fractions[1] * x_width)
    x3 = int(xmin + fractions[2] * x_width)

    y1 = np.nanargmax(edges[ymin:ymax, x1]) + ymin
    y2 = np.nanargmax(edges[ymin:ymax, x2]) + ymin
    y3 = np.nanargmax(edges[ymin:ymax, x3]) + ymin
    r1sq = x1 ** 2 + y1 ** 2
    r2sq = x2 ** 2 + y2 ** 2
    r3sq = x3 ** 2 + y3 ** 2
    A = np.array([[0, 0, 0, 1],
                  [r1sq, x1, y1, 1],
                  [r2sq, x2, y2, 1],
                  [r3sq, x3, y3, 1]])
    m11 = getMatrixMinor(A, 1, 1)
    m12 = getMatrixMinor(A, 1, 2)
    m13 = getMatrixMinor(A, 1, 3)
    m14 = getMatrixMinor(A, 1, 4)
    px = _correct_pixel(0.5 * m12 / m11)
    py = _correct_pixel(-0.5 * m13 / m11)
    r = np.sqrt(px ** 2 + py ** 2 + m14 / m11)
    return r, px, py, (x1, x2, x3), (y1, y2, y3)


def getMatrixMinor(m, i, j):
    # ones indexed
    row = np.arange(1, 5)
    col = np.arange(1, 5)

    minor = np.array([r[col != j] for r in m[row != i]])
    return np.linalg.det(minor)
