from typing import Union, Tuple

import numpy as np
import xarray as xr

from altius_l1pp_inflight.srf.SrfInterface import SrfInterface


class GaussianSrf(SrfInterface):
    """
    Simple implementation of Gaussian SRF with varying FWHM in wavelength.
    """
    LINEAR_INTERP = 'linear'
    LOG_INTERP = 'log'

    def __init__(self, grid_wavelengths, grid_fwhm):
        self._grid_wavelengths = grid_wavelengths
        self._grid_fwhm = grid_fwhm
        self._wavindex: np.ndarray = np.argsort(self._grid_wavelengths)

    def get_srf(self, center_wavelength_nm, new_wavelengths=None,
                interp_method=LINEAR_INTERP, normalize=True) -> Tuple[xr.DataArray, float]:
        """
        Returns the SRF closest to the specified wavelength, and the actual wavelength

        Parameters
        ----------
        center_wavelength_nm : float
            Center wavelength to obtain the SRF from the tunable wavelengths

        new_wavelengths : np.ndarray, Optional
            The grid of wavelengths, in nm, to evaluate/interpolate the SRF over.  If not
            specified, will return the SRF at the original measured wavelengths in wavelength_grid

        interp_method : str, Optional
            'linear' or 'log'.  Interpolation method for evaluating the SRF over new_wavelengths

        normalize : bool, Optional
            If True, normalizes the integral of the SRF to be 1.0.
            If False, the values of the SRF are adjusted so that the integral matches the original
            data's integral

        Returns
        -------
        +srf : xr.DataArray
            The SRF with tuned wavelength near to center_wavelength_nm.  x-coordinate 'wavelength' set
            by the new_wavelengths parameter
        actual_wavelength : float
            The nearest tunable wavelength to the input center_wavelength_nm
        """

        idx0, idx1 = self._find_bounding_wavelength_indices(center_wavelength_nm)
        fwhm0 = self._grid_fwhm[idx0]
        fwhm1 = self._grid_fwhm[idx1]
        # Linearly interpolate the FWHM:
        fwhm = np.interp(center_wavelength_nm, [self._grid_wavelengths[idx0], self._grid_wavelengths[idx1]],
                         [fwhm0, fwhm1])
        sigma = fwhm / (2 * np.sqrt(2 * np.log(2)))
        srf_wavelen = np.arange(center_wavelength_nm - sigma * 6,
                                center_wavelength_nm + sigma * 6,
                                step=0.01)  # 0.01 nm res

        srf = np.exp(-(srf_wavelen - center_wavelength_nm) ** 2 / (2 * sigma ** 2))
        srf /= np.sum(srf)
        srf = xr.DataArray(data=srf, coords={'wavelength': srf_wavelen}, dims={'wavelength'})

        if new_wavelengths is not None:
            srf = srf.interp(wavelength=new_wavelengths)
            srf /= np.nansum(srf.values)

        return srf, center_wavelength_nm

    def convolve_spectrum_with_srf(self, convolution_wavelengths: Union[float, np.ndarray],
                                   spectrum_wavelengths, spectrum) -> np.ndarray:
        """
            Convolves the spectrum with the SRF at each wavelength.

            Parameters
            ----------
            spectrum_wavelengths : np.ndarray
                Wavelengths in nm of the input spectrum

            spectrum : np.ndarray
                Input high-resolution spectrum to be convolved with the corrected SRF

            convolution_wavelengths : Union[float, np.ndarray]
                Wavelengths (in nm) to evaluate the resulting convolution at

            Returns
            -------
            convolved_spectrum : np.ndarray
                The resulting spectrum convolved with SRFs at each convolution wavelength
        """

        convolved_spectrum = np.zeros((len(convolution_wavelengths), 1))
        for k, w in enumerate(convolution_wavelengths):
            srf_w, _ = self.get_srf(w, spectrum_wavelengths)
            convolved_spectrum[k] = np.nansum(srf_w.values * spectrum)

        return convolved_spectrum

    def _find_bounding_wavelength_indices(self, wavelength_nm):
        idx1 = np.searchsorted(self._grid_wavelengths, wavelength_nm, sorter=self._wavindex)
        if idx1 == 0:
            idx1 = 1
        if idx1 == self._grid_wavelengths.size:
            idx1 = self._grid_wavelengths.size - 1
        idx0 = idx1 - 1
        return idx0, idx1

    def get_wavelength_grid(self, resolution=None):
        if resolution is None:
            return self._grid_wavelengths

        return np.arange(self._grid_wavelengths[0],
                         self._grid_wavelengths[-1],
                         step=resolution)
