import matplotlib.font_manager as font_manager
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
from sasktran.solarspectrum import SolarSpectrum
import os

# Set the font dictionaries (for plot title and axis titles)
from altius_l1pp_inflight.srf import wavelen_registration, uv_srf

title_font = {'name': 'Arial', 'size': '20', 'color': 'black', 'weight': 'normal',
              'verticalalignment': 'bottom'}  # Bottom vertical alignment for more space
axis_font = {'name': 'Arial', 'size': '18'}
# Set the font properties (for use in legend)
font_path = 'C:\\Windows\\Fonts\\Arial.ttf'
font_prop = font_manager.FontProperties(fname=font_path, size=16)

directory = 'A:\\USASK\\ALTIUS_DATA\\l1pp_datafiles\\+srf'

srf = uv_srf.AltiusUVSrf(directory)

solar_spectrum = SolarSpectrum(name='SAO2010')
wavelength_grid = srf.get_wavelength_grid()

sao = solar_spectrum.irradiance(wavelength_grid)
solar = xr.DataArray(sao, dims={'wavelength'}, coords={'wavelength': wavelength_grid})

assume_gaussian = True
WL = np.linspace(250, 350, 1001)
fit_fwhm, actual_fwhm, WL, sim_spectrum, fit_spectrum, rel_percent_diff = \
    wavelen_registration.fit_using_gaussian_approximation(srf, WL, assume_gaussian, snr=None, solar_distance_AU=1.0)
# solar_distance_AU=0.9893723729948187)


# fig1, ax1 = plt.subplots()
# # ax1.plot(+srf.tunable_wavelengths(), fwhm_sensitivity)
# idx280 = np.abs(srf_sensitivity.wavelengths - 280).argmin()
# idx300 = np.abs(srf_sensitivity.wavelengths - 300).argmin()
# idx315 = np.abs(srf_sensitivity.wavelengths - 315).argmin()
# idx335 = np.abs(srf_sensitivity.wavelengths - 335).argmin()
# ax1.plot(srf_sensitivity.fwhm_pert_hr, srf_sensitivity.delta_irradiance[idx280])
# ax1.plot(srf_sensitivity.fwhm_pert_hr, srf_sensitivity.delta_irradiance[idx300])
# ax1.plot(srf_sensitivity.fwhm_pert_hr, srf_sensitivity.delta_irradiance[idx315])
# ax1.plot(srf_sensitivity.fwhm_pert_hr, srf_sensitivity.delta_irradiance[idx335])
# plt.xlabel('Wavelength [nm]', **axis_font)
# plt.ylabel('[Irradiance]/nm')
# plt.title('Sensitivity to Central Lobe Perturbations at Different Wavelengths', **title_font)
# plt.legend(['280 nm', '300 nm', '315 nm', '335 nm'])
# plt.xticks(**axis_font)
# plt.yticks(**axis_font)

gaussian_fit_spectrum = 0 * fit_fwhm
wl_hr = np.linspace(WL[0] - 20, WL[-1] + 20, int((WL[-1] - WL[0] + 40) / 0.04))
solar_hr = SolarSpectrum().irradiance(wl_hr)
# srf_fit_spectrum = 0 * fit_fwhm

for k, w in enumerate(WL):
    sigma = wavelen_registration.fwhm_2_sigma(actual_fwhm.values[k])
    g = np.exp(-0.5 * (wl_hr - w) ** 2 / sigma ** 2)
    g /= np.sum(g)
    sigmaf = wavelen_registration.fwhm_2_sigma(fit_fwhm[k])
    gf = np.exp(-0.5 * (wl_hr - w) ** 2 / sigmaf ** 2)
    if ~assume_gaussian:
        gf = sim_spectrum.srf.center_lobe_adjusted_fwhm(w, wl_hr, fit_fwhm[k])
    gf /= np.sum(gf)
    gaussian_fit_spectrum[k] = np.sum(gf * solar_hr)
    # s = sim_spectrum.+srf.center_lobe_adjusted_fwhm(w, wl_hr, fit_fwhm[k])
    # s /= np.sum(s)
    # srf_fit_spectrum[k] = np.sum(s * solar_hr)

fig2, ax2 = plt.subplots()
ax2.plot(WL, fit_fwhm)
actual_fwhm.plot(ax=ax2)
plt.xlabel('Wavelength [nm]', **axis_font)
plt.ylabel('FWHM [nm]')
plt.title('Fit vs. Actual FWHM', **title_font)
plt.legend(['Fit FWHM', 'Actual FWHM'], prop=font_prop)
plt.xticks(**axis_font)
plt.yticks(**axis_font)

fig3, ax3 = plt.subplots()
cor_obs = sim_spectrum.blurred_spectrum(sim_spectrum.SIDE_LOBE_REMOVED)
cor_obs.sel(wavelength=WL).plot(ax=ax3)
observed_spectrum = sim_spectrum.blurred_spectrum(sim_spectrum.FULL_SRF)
observed_spectrum.sel(wavelength=WL).plot(ax=ax3)
# ax3.plot(WL, gaussian_fit_spectrum)
ax3.plot(WL, fit_spectrum)
# ax3.plot(WL, srf_fit_spectrum)
plt.xlabel('Wavelength [nm]', **axis_font)
plt.ylabel('Irradiance')
plt.title('Comparison of Corrected Observation and Resulting Fit', **title_font)
plt.legend(['Corrected Observation', 'Uncorrected Observation', 'Fit Gaussian Convolved'], prop=font_prop)
plt.xticks(**axis_font)
plt.yticks(**axis_font)

fig4, ax4 = plt.subplots()
srf_263_5, _ = sim_spectrum.srf.get_srf(263.5)
wl = sim_spectrum.srf.get_base_wavelength_measurement_grid()
g_263_5, _, _ = sim_spectrum.srf.get_gaussian(263.5, wl)
srf_263_6, _ = sim_spectrum.srf.get_srf(263.6)
g_263_6, _, _ = sim_spectrum.srf.get_gaussian(263.6, wl)
plt.plot(wl, srf_263_5)
plt.plot(wl, g_263_5, linestyle='dashed')
plt.plot(wl, srf_263_6)
plt.plot(wl, g_263_6, linestyle='dashed')
plt.legend(['263.5 nm SRF', '263.5 nm Gaussian', '263.6 nm SRF', '263.6 nm Gaussian'], prop=font_prop)
plt.xlabel('Wavelength [nm]', **axis_font)
plt.ylabel('Irradiance')
plt.title('Comparison of non-Gaussian SRFs with Gaussians', **title_font)
plt.xticks(**axis_font)
plt.yticks(**axis_font)

fig5, ax5 = plt.subplots()
diff_fit_corrected = 100 * (gaussian_fit_spectrum - cor_obs.sel(wavelength=WL)) / cor_obs.sel(wavelength=WL)

# diff_fit_srf_corrected = (srf_fit_spectrum - corrected_observation.sel(wavelength=WL)) / \
#                          corrected_observation.sel(wavelength=WL)
# diff_fit_corrected.plot(ax=ax5)
# diff_fit_srf_corrected.plot(ax=ax5)
ax5.plot(WL, rel_percent_diff)
# plt.legend(['Diff Using Gaussian Shape FWHM', 'Diff Using SRF Shape FWHM'], prop=font_prop)
ax5.plot([WL[0], WL[-1]], [0.1, 0.1], color='black', linestyle='dashed', linewidth=2.5)
ax5.plot([WL[0], WL[-1]], [-0.1, -0.1], color='black', linestyle='dashed', linewidth=2.5)
plt.xlabel('Wavelength [nm]', **axis_font)
plt.ylabel('Percent Difference')
plt.title('Comparison of Relative Difference of Fit to Corrected Spectrum', **title_font)
plt.xticks(**axis_font)
plt.yticks(**axis_font)
plt.ylim([-2, 2])

fig6, ax6 = plt.subplots()
a_f = actual_fwhm.sel(wavelength=WL)
actual_fwhm_wavenumber = a_f.values * 1e-7 * (1 / (a_f.wavelength.values * 1e-7)) ** 2
wavenum_error = 0.05 * actual_fwhm_wavenumber
fit_fwhm_wavenumber = fit_fwhm * 1e-7 * (1 / (WL * 1e-7)) ** 2
plt.plot(a_f.wavelength.values, actual_fwhm_wavenumber, linewidth=0.8)
plt.plot(WL, fit_fwhm_wavenumber, linewidth=0.8)
plt.legend(['Actual FWHM with 5% error', 'Fit FWHM'], prop=font_prop)
plt.fill_between(a_f.wavelength.values, actual_fwhm_wavenumber - wavenum_error, actual_fwhm_wavenumber + wavenum_error,
                 facecolor='blue', alpha=0.3)
plt.xlabel('Wavelength [nm]', **axis_font)
plt.ylabel('FWHM [cm-1]')
plt.title('Comparison of Fit FWHM to Actual FWHM with SRD', **title_font)
plt.xticks(**axis_font)
plt.yticks(**axis_font)

fig7, ax7 = plt.subplots()
fwhm_perc_diff = 100 * (fit_fwhm_wavenumber - actual_fwhm_wavenumber) / actual_fwhm_wavenumber
ax7.plot(WL, fwhm_perc_diff)
plt.xlabel('Operational Wavelength [nm]', **axis_font)
plt.ylabel('FWHM [cm-1]')
plt.title('Percent difference of Fit FWHM Relative to Actual', **title_font)
plt.xticks(**axis_font)
plt.yticks(**axis_font)
ax7.plot(WL, 5 * np.ones(len(WL)), color='black', linestyle='dashed')
ax7.plot(WL, -5 * np.ones(len(WL)), color='black', linestyle='dashed')

valid_wavelengths = WL[np.argwhere(np.abs(fwhm_perc_diff) < 5.0)]

outdir = 'A:\\USASK\\ALTIUS_DATA\\altius_wavelength_registration'

with open(os.path.join(outdir, 'valid_uv_wavelengths.txt'), 'w') as f:
    for w in valid_wavelengths:
        f.write('%f\n' % w)

plt.show()
