import os

import matplotlib.font_manager as font_manager
import matplotlib.pyplot as plt
import numpy as np
import progressbar
import sktime.mjd as mjd_conv
import xarray as xr
from sasktran import SolarSpectrum

from altius_l1pp_inflight.configuration import altius_config
from altius_l1pp_inflight.scene.altius_scene import SpsToUsaskSolarSceneInterface
from altius_l1pp_inflight.scene.disk_geometry import D_1AU, SolarGeometry
from altius_l1pp_inflight.scene.generation.solar.hmi_files import HmiFileManager, SolarDiskRadianceResult, \
    HmiSunspotScenario
from altius_l1pp_inflight.scene.generation.solar.interpolate_solar_disk import SimulatedSolarDiskInterpolator
from altius_l1pp_inflight.scene.generation.solar.virtual_solar_detector import RasterScanGrid, VirtualDetectorPixelArray
from altius_l1pp_inflight.scene.limbdark_mu_pierceandslaughter1977 import LimbDarkening_PS1977
from altius_l1pp_inflight.scene.util import convert_to_DN_and_add_noise
from altius_l1pp_inflight.srf.blurred_spectra import SAO2010
from altius_l1pp_inflight.srf.solar.srf_util import compute_irradiance_center_pixel_scaling_factor
from altius_l1pp_inflight.srf.uv_srf import AltiusUVSrf, FwhmPerturbations
from altius_l1pp_inflight.srf.wavelen_registration import fwhm_2_sigma, fit_measured_spectrum_window_least_squares, \
    fit_fwhm_to_corrected_spectrum, SimulatedUVSolarSpectrum

"""
Top level class to determine the wavelength registration for an ALTIUS image.
It is assumed that the ALTIUS image has been corrected for all major effects,
though the PRNU correction will have the most recent update.

A group of pixels in the center of the solar disk is used as the measurement for
modelling the solar irradiance.  This is repeated for various locations of the
sun on the detector.

The observed solar spectrum is "corrected" assuming perfect knowledge of the SRF.
The contribution attributed to the side band of the SRF at each operational wavelength
to an ideal solar spectrum is determined, then subtracted from the measured spectrum.
To restore the summation to the proper scale, the result is divided by the area under
the central lobe at each operational wavelength.

From here, a fitting procedure where a Gaussian SRF is adjusted to fit the corrected
spectrum at each measurement wavelength ensues.

This is then repeated for a number of constant wavelength shifts from roughly -1 to +1 nm.
The ideal wavelength shift-FWHM combo is then determined by TBD
"""

# Set the font dictionaries (for plot title and axis titles)
title_font = {'name': 'Arial', 'size': '20', 'color': 'black', 'weight': 'normal',
              'verticalalignment': 'bottom'}  # Bottom vertical alignment for more space
axis_font = {'name': 'Arial', 'size': '18'}
# Set the font properties (for use in legend)
font_path = 'C:\\Windows\\Fonts\\Arial.ttf'
font_prop = font_manager.FontProperties(fname=font_path, size=16)
config_filename = "globalkey_3x3_grid.yaml"
recompute = True

config = altius_config.AltiusL1ppInflightConfig(config_filename)

srf_directory = altius_config.get_srf_directory()

altius_srf = AltiusUVSrf(srf_directory, fwhm_method_using_area=True)

home_dir = config.get_processed_wavelength_reg_file_directory()
output_directory = os.path.join(home_dir, 'altius_observation', 'center_detector')

nc_file = 'altius_wavelength_uv_srf_calibration.nc'
nc_file = os.path.join(output_directory, nc_file)


def run_least_squares_stretch_and_shift_fit(observed_center_irradiance: xr.DataArray,
                                            hr_wavelengths: np.ndarray,
                                            irradiance_scaling_factor: np.ndarray):
    """
        One of the two main ways we can calibrate the SRF at a chosen wavelength.
        This fit takes a range of wavelengths around some target wavelength and applies
        a least squares fit to the spectrum to obtain an FWHM and wavelength shift
    """
    scaling_factor = xr.DataArray(irradiance_scaling_factor, dims={'wavelength'}, coords={'wavelength': hr_wavelengths})
    # perform the least squares fit - all wavelengths are intended for the single analysis wavelength at the center of the window
    params = np.r_[fwhm_2_sigma(2.5), 0.0]
    sigma = fwhm_2_sigma(np.arange(0.5, 5, step=0.1))
    # Important: scaling factor has to be across larger wavelength than fit window
    FIT_WINDOW = 10
    wavelengths = observed_center_irradiance.wavelength.values
    solar = SAO2010(sigma, start_wavelen=wavelengths[0] - FIT_WINDOW, end_wavelen=wavelengths[-1] + FIT_WINDOW,
                    solar_distance_AU=average_solar_distance, scaling_factor=scaling_factor)

    fit_results = fit_measured_spectrum_window_least_squares(wavelengths,
                                                             observed_center_irradiance.values,
                                                             params,
                                                             solar)
    return fit_results


def run_fwhm_fit(wavelengths, observed_side_lobes_removed, irradiance_scaling_factor, hr_wavelengths):
    """
        One of the two main ways we can calibrate the SRF at a chosen wavelength.
        This fit assumes the wavelength shift is zero and tries to match the solar
        spectrum by altering the FWHM until the spectra match
    """
    fit_fwhm = np.zeros_like(wavelengths)
    rel_percent_diff = np.zeros_like(wavelengths)
    fit_spectrum = np.zeros_like(wavelengths)
    bar = progressbar.ProgressBar(min_value=0, max_value=len(wavelengths))
    fwhm_perturb = FwhmPerturbations(srf=altius_srf, solar_distance_AU=average_solar_distance / D_1AU,
                                     scaling_factor=irradiance_scaling_factor,
                                     hr_wavelengths=hr_wavelengths)
    fwhm_step_function = fwhm_perturb.step_fwhm_assume_gaussian
    for k, wl in enumerate(wavelengths):
        corrected_val = observed_side_lobes_removed[k]
        delta_spectrum, fwhm_guess, rel_spectral_delta = \
            fit_fwhm_to_corrected_spectrum(wl, corrected_val, fwhm_step_function)
        rel_percent_diff[k] = 100 * rel_spectral_delta
        fit_spectrum[k] = corrected_val - delta_spectrum
        fit_fwhm[k] = fwhm_guess
        bar.update(k)

    return fit_fwhm, rel_percent_diff, fit_spectrum


def compute_altius_images_and_save():
    scene_interface = SpsToUsaskSolarSceneInterface()

    pixel_dim = altius_config.get_detector_pixel_dimensions(350, 'native', config_filename=config_filename)
    num_pixels_vertical = pixel_dim[0]
    num_pixels_horizontal = pixel_dim[1]
    raster_index = 4  # places the sun at the center
    num_vert_raster = altius_config.get_num_vertical_raster_scans(config_filename=config_filename)
    num_horz_raster = altius_config.get_num_horizontal_raster_scans(config_filename=config_filename)
    detector_angular_width = altius_config.get_detector_fov_from_wavelength(350, config_filename=config_filename)
    maxDN = altius_config.get_max_DN(config_filename)
    max_photons_s_cm2 = altius_config.get_max_radiance(config_filename)
    fov = altius_config.get_detector_fov_from_wavelength(300, config_filename)
    factor = maxDN / max_photons_s_cm2
    file_manager = HmiFileManager(config)
    results = file_manager.load_scenario_files_as_images(HmiSunspotScenario.NO_SUNSPOTS)
    all_scaled_images = []
    for res in results:
        if not isinstance(res, SolarDiskRadianceResult):
            raise AssertionError('Object is not of type HmiSolarDiskRadianceResult')
        all_scaled_images.append(res)
    all_scaled_images = np.array(all_scaled_images)
    solar_disk_interpolator = SimulatedSolarDiskInterpolator(all_scaled_images)
    solar_pixel_radius = results[0].average_solar_radius_in_pixels()
    solar_angular_radius = results[0].angular_radius_degrees()
    solar_scan_grid = RasterScanGrid(solar_angular_radius,
                                     0.1, 0.1, detector_angular_width[0],
                                     detector_angular_width[1],
                                     num_vert_raster, num_horz_raster)
    vdpa = VirtualDetectorPixelArray(disk_radius_pixels=solar_pixel_radius,
                                     solar_scan_grid=solar_scan_grid,
                                     boresight_horz_portion=1.0,
                                     boresight_vert_portion=1.0)
    # Example disjointed wavelengths
    wavelength_grid = np.linspace(290, 380, num=8001)
    # WL = np.union1d(np.linspace(310, 315, 25), np.linspace(320, 340, 40))

    # WL = np.loadtxt(os.path.join(os.environ['ALTIUS_DATABASE_FOLDER'],
    #                        'altius_wavelength_registration',
    #                        'valid_uv_wavelengths.txt'))
    WL = np.linspace(295, 305, 10)
    calibration_wavelength = 300

    t0 = solar_disk_interpolator.grid_mjd_times()[0]
    tend = solar_disk_interpolator.grid_mjd_times()[-1]
    TIMES = np.linspace(t0, tend, len(WL))

    boresight = (num_pixels_vertical / 2, num_pixels_horizontal / 2)

    scene_interface.open_file_for_write(nc_file,
                                        calibration_wavelength,
                                        (fov[0], fov[1]),
                                        (num_pixels_vertical, num_pixels_horizontal),
                                        boresight)

    for wavelength, time in zip(WL, TIMES):
        datetime_val = mjd_conv.mjd_to_datetime(time)
        # filepath = os.path.join(output_directory, 'altius_obs_' +
        #                         '{:.2f}'.format(wavelength).replace('.', '_') + '__' +
        #                         datetime_val.strftime('%m_%d_%Y__%H%M%S') + '.pickle')

        print('Computing ALTIUS +scene for ' + '{:.2f}'.format(wavelength) +
              ' nm and ' + datetime_val.strftime('%m_%d_%Y, %H:%M:%S'))
        # +srf, actual_wl = altius_srf.get_srf(wavelength, wavelength_grid, normalize=False)

        solar_image = solar_disk_interpolator.interpolate_disk(datetime_val)

        # photons/s/cm2/nm/sr:
        scaled_image = solar_image.to_scaled_image_result(wavelength=wavelength,
                                                          srf=altius_srf,
                                                          horizontal_pixel_shift=0,
                                                          vertical_pixel_shift=0)
        vdpa.set_current_raster_image(scaled_image, num_pixels_vertical,
                                      num_pixels_horizontal, raster_index)
        image, image_geometry = vdpa.to_altius_resolution_image(True)

        image_dn = convert_to_DN_and_add_noise(image, factor)

        expected_altius_solar_center_horz = image_geometry['altius_solar_center_horz']
        expected_altius_solar_center_vert = image_geometry['altius_solar_center_vert']
        expected_altius_solar_radius = image_geometry['altius_solar_radius']
        sat_to_sun_distance_km = solar_image.actual_solar_distance()

        scene_interface.write_next_solar_scene_to_netcdf(mjd_conv.mjd_to_datetime64(time),
                                                         image_dn,
                                                         wavelength,
                                                         (expected_altius_solar_center_vert,
                                                          expected_altius_solar_center_horz),
                                                         expected_altius_solar_radius,
                                                         sat_to_sun_distance_km,
                                                         np.array([1, 0, 0]),
                                                         np.array([1, 0, 0]),
                                                         0.1)

    scene_interface.close_write_nc()


if recompute:
    compute_altius_images_and_save()

scene_interface = SpsToUsaskSolarSceneInterface()
scene_interface.open_read_dataset(nc_file)

n_images = scene_interface.get_num_scenes()

# files = [os.path.join(output_directory, f) for f in os.listdir(output_directory) if os.path.isfile(os.path.join(output_directory, f))]
wavelengths = np.zeros((n_images,))
observed_radiance = np.zeros_like(wavelengths)

maxDN = altius_config.get_max_DN(config_filename)
max_photons_s_cm2 = altius_config.get_max_radiance(config_filename)
dn_to_photons_factor = max_photons_s_cm2 / maxDN

average_solar_distance = 0
solid_angle_solar_disk = 0
solid_angle_center_pixels = 0

pixel_dim = config.get_detector_pixel_dimensions('uv', 'native')
num_pixels_vertical = pixel_dim[0]
num_pixels_horizontal = pixel_dim[1]

h = np.arange(0, num_pixels_horizontal) + 0.5
v = np.arange(0, num_pixels_vertical) + 0.5

vv, hh = np.meshgrid(h, v)
center_radius = 6
mu_start_avg = 0

limb_darkening_model = LimbDarkening_PS1977()
average_radius_pixels = 0
average_vert_shift = 0
average_horz_shift = 0
average_center_horz = 0
average_center_vert = 0
avg_boresight_horizontal = 0
avg_boresight_vertical = 0

bv, bh = config.get_pixel_boresights(channel='uv', mode='native')

for idx in range(n_images):
    scene = scene_interface.load_scene(idx)
    image = scene.get_altius_image()
    vc = scene.get_expected_altius_solar_center_vert()
    hc = scene.get_expected_altius_solar_center_horz()
    average_solar_distance += scene.get_satellite_to_moon_distance_km() / n_images

    pixel_radii = np.sqrt((vv - vc) ** 2 + (hh - hc) ** 2)

    mask = pixel_radii <= center_radius  # 6 pixel radius
    center_pixels = image[mask]  # DN -> photons/s/cm2/nm/sr
    solar_geometry = SolarGeometry(len(image[0]), len(image), bv, bh)
    solar_geometry.set_disk(scene.get_expected_altius_solar_radius(),
                            scene.get_expected_altius_solar_center_horz(),
                            scene.get_expected_altius_solar_center_vert(),
                            scene.get_satellite_to_sun_distance_km())
    delta_omega = solar_geometry.delta_omega_view()

    # limb darkening correction:
    disk_radius = scene.get_expected_altius_solar_radius()
    average_radius_pixels += disk_radius / n_images
    mu_start = np.sqrt(1 - (center_radius / disk_radius) ** 2)
    mu_start_avg += mu_start / n_images

    # TODO apply centering to find the vertical and horizontal center components
    horz_center = scene.get_expected_altius_solar_center_horz()
    vert_center = scene.get_expected_altius_solar_center_vert()

    irrad, _ = scene.compute_irradiance(horz_center, vert_center, dn_to_photons_factor,
                                        center_radius)  # photons/cm2/s/sr/nm
    wavelengths[idx] = scene.get_wavelength()
    observed_radiance[idx] = irrad
    solid_angle_solar_disk += np.nansum(delta_omega[pixel_radii <= disk_radius]) / n_images
    solid_angle_center_pixels += np.nansum(delta_omega[mask]) / n_images

    boresight = scene.get_boresight()

    average_center_horz += hc / n_images
    average_center_vert += vc / n_images
    avg_boresight_horizontal += boresight[1] / n_images
    avg_boresight_vertical += boresight[0] / n_images

wavelengths = wavelengths.flatten()
observed_radiance = xr.DataArray(observed_radiance.flatten(), dims={'wavelength'},
                                 coords={'wavelength': wavelengths})

fig, ax = plt.subplots()
observed_radiance.plot(ax=ax)
plt.show()

# TODO add noise to scenes
# TODO get irradiance from center pixels
# TODO incorporate wavelength shift (iterative...?)

solar_spectrum = SolarSpectrum()
hr_wavelengths = np.linspace(250, 380, 6500)  # 0.02 nm resolution

irradiance_scaling_factor = compute_irradiance_center_pixel_scaling_factor(hr_wavelengths,
                                                                           average_radius_pixels,
                                                                           center_radius,
                                                                           average_solar_distance,
                                                                           average_center_horz,
                                                                           average_center_vert,
                                                                           num_pixels_horizontal,
                                                                           num_pixels_vertical,
                                                                           avg_boresight_horizontal,
                                                                           avg_boresight_vertical)

hr_solar_irradiance = solar_spectrum.irradiance(hr_wavelengths,
                                                solardistance=average_solar_distance / D_1AU)
# hr_solar_irradiance /= (hr_wavelengths[1] - hr_wavelengths[0]) # photons/cm2/sec/nm
irradiance_3nm = solar_spectrum.irradiance(hr_wavelengths,
                                           solardistance=average_solar_distance / D_1AU,
                                           fwhm=3.0)

hr_solar_irradiance *= irradiance_scaling_factor
irradiance_3nm *= irradiance_scaling_factor

wavelengths = wavelengths.flatten()

observed_side_lobes_removed, central_lobe_portion_corr, central_lobe_fwhm_corr, srf_central_lobe_sum, srf_side_lobe_sum = \
    altius_srf.convolve_corrected_srf(hr_wavelengths, hr_solar_irradiance, wavelengths,
                                      observed_radiance)  # photons/cm2/sec/sr

plt.plot(hr_wavelengths, hr_solar_irradiance, alpha=0.3, linewidth=0.8)
plt.plot(wavelengths, observed_radiance)
plt.plot(wavelengths, observed_side_lobes_removed)
plt.plot(hr_wavelengths, irradiance_3nm)
plt.legend(['HiRes SAO2010', 'Observed Spectrum', 'Observed Corrected Spectrum', 'SAO2010 3 nm Gaussian'])
plt.xlabel('Wavelength (nm)')
plt.ylabel('Radiance (photons/cm2/sec/nm/sr)')
plt.show()

fit_fwhm, rel_percent_diff, fit_spectrum = run_fwhm_fit(wavelengths, observed_side_lobes_removed,
                                                        irradiance_scaling_factor, hr_wavelengths)

sim_spectrum = SimulatedUVSolarSpectrum(altius_srf, None,
                                        solar_distance_AU=average_solar_distance / D_1AU)  # trying to fit this with a shifted and blurred solar spectrum with only Gaussian

fig1, ax1 = plt.subplots()
actual_fwhm = sim_spectrum.central_lobe_fwhm().interp(wavelength=wavelengths)
actual_fwhm.plot(ax=ax1)
ax1.plot(wavelengths, fit_fwhm)
plt.legend(['Actual FWHM', 'Fit FWHM'])
plt.xlabel('Wavelength (nm)')
plt.ylabel('FWHM (nm)')

fig5, ax5 = plt.subplots()
ax5.plot(wavelengths, rel_percent_diff)
# plt.legend(['Diff Using Gaussian Shape FWHM', 'Diff Using SRF Shape FWHM'], prop=font_prop)
ax5.plot([wavelengths[0], wavelengths[-1]], [0.1, 0.1], color='black', linestyle='dashed', linewidth=2.5)
ax5.plot([wavelengths[0], wavelengths[-1]], [-0.1, -0.1], color='black', linestyle='dashed', linewidth=2.5)
plt.xlabel('Wavelength [nm]', **axis_font)
plt.ylabel('Percent Difference')
plt.title('Comparison of Relative Difference of Fit to Corrected Spectrum', **title_font)
plt.xticks(**axis_font)
plt.yticks(**axis_font)
plt.ylim([-2, 2])

fig6, ax6 = plt.subplots()
a_f = actual_fwhm.sel(wavelength=wavelengths)
actual_fwhm_wavenumber = a_f.values * 1e-7 * (1 / (a_f.wavelength.values * 1e-7)) ** 2
wavenum_error = 0.05 * actual_fwhm_wavenumber
fit_fwhm_wavenumber = fit_fwhm * 1e-7 * (1 / (wavelengths * 1e-7)) ** 2
plt.plot(a_f.wavelength.values, actual_fwhm_wavenumber, linewidth=0.8)
plt.plot(wavelengths, fit_fwhm_wavenumber, linewidth=0.8)
plt.legend(['Actual FWHM with 5% error', 'Fit FWHM'], prop=font_prop)
plt.fill_between(a_f.wavelength.values, actual_fwhm_wavenumber - wavenum_error, actual_fwhm_wavenumber + wavenum_error,
                 facecolor='blue', alpha=0.3)
plt.xlabel('Wavelength [nm]', **axis_font)
plt.ylabel('FWHM [cm-1]')
plt.title('Comparison of Fit FWHM to Actual FWHM with SRD', **title_font)
plt.xticks(**axis_font)
plt.yticks(**axis_font)

fig2, ax2 = plt.subplots()
ax2.plot(wavelengths, sim_spectrum.blurred_spectrum(mode=sim_spectrum.SIDE_LOBE_REMOVED, wavelengths=wavelengths))
ax2.plot(wavelengths, fit_spectrum)
plt.legend(['Simulated observed spectrum', 'Fit Spectrum'])
plt.show()
