import csv
import os
import sys
from typing import Tuple

import numpy as np
import progressbar.bar as progress_bar
import scipy.io
import xarray as xr
from sasktran.solarspectrum import SolarSpectrum
from scipy.interpolate import RectBivariateSpline
from scipy.optimize import fsolve as find_root
from scipy.stats import norm as norm_dist

from altius_l1pp_inflight.srf.SrfInterface import SrfInterface

RESOLUTION = 0.04


def load_decimated_srf():
    csv.field_size_limit(int(sys.maxsize / 10000000000))
    srf = []
    with open('srf_uv.txt') as f:
        rd = csv.reader(f)
        for i, row in enumerate(rd):
            line = []
            if (i - 1) % 10 == 0 or i == 0:
                x = row[0].split()
                for v in x:
                    line.append(float(v))
                srf.append(line)

    srf = np.array(srf)

    optics = np.genfromtxt('uv_optics_srf_11.txt', dtype=np.float, skip_header=1)

    # interpolate optics WL (column 0) BL_BOL (column 1) onto +srf grid (column 0)
    optics_filter = np.interp(srf[0, 5:], optics[:, 0], optics[:, 1], right=0, left=0)
    optics_filter = optics_filter[:, np.newaxis]

    result = dict()
    result['center_wavelength'] = srf[1:, 0]
    result['wavelength_grid'] = srf[0, 5:]
    result['+srf'] = srf[1:, 5:] * optics_filter.T

    return result


def test_central_lobe_portion_gaussian():
    wavelen_grid = np.linspace(202, 400, 19800)  # 0.01 nm resolution
    central_lobe_wl = np.linspace(250, 355, 1051)  # 0.1 nm resolution
    s = SolarSpectrum()
    solar = s.irradiance(wavelen_grid)
    solar = solar[:, np.newaxis]
    sigma = 3 / (2 * np.sqrt(2 * (np.log(2))))
    wg = np.tile(wavelen_grid, [1051, 1])
    w = np.tile(central_lobe_wl, [19800, 1]).T
    g = 1 / np.sqrt(2 * np.pi * sigma ** 2) * np.exp(-(wg - w) ** 2 / (2 * sigma ** 2))
    indiv_conv_3nm_fwhm = g * solar.T
    idx = np.abs(w - wg).argmin(axis=1)
    portion = [np.sum(indiv_conv_3nm_fwhm[k, i - 300:i + 300]) / np.sum(indiv_conv_3nm_fwhm[k, :])
               for k, i in enumerate(idx)]  # +/- 300 idx == +/-3 nm at 0.01 nm resolution

    lobe_portion = xr.DataArray(portion, dims={'wavelength'}, coords={'wavelength': central_lobe_wl},
                                attrs={'long_name': 'Ratio of Central lobe to Entire Spectrum'})
    return g, lobe_portion


def make_gaussian(center_wavelength, fwhm, wavelengths):
    sigma = fwhm_2_sigma(fwhm)
    g = np.exp(-0.5 * (wavelengths - center_wavelength) ** 2 / sigma ** 2)
    g /= np.sum(g)
    return g


FWHM_CONVERT = 2 * np.sqrt(2 * np.log(2))


def fwhm_2_sigma(fwhm):
    return fwhm / FWHM_CONVERT


def sigma_2_fwhm(sigma):
    return FWHM_CONVERT * sigma


# TODO not sure if this is actually used anywhere
class SrfSensitivity:
    """
    Class for computing the sensitivity of the solar spectrum to changes in the FWHM of a simple
    Gaussian spectral PSF.  Can be used in determining the optimal change in the FWHM to fit a Gaussian
    to an observed solar spectrum that has been convolved with an SRF
    """

    def __init__(self, fwhm_perturbations, wavelengths, delta_irradiance):
        self.fwhm_perturbations = fwhm_perturbations
        self.wavelengths = wavelengths
        # self.delta_irradiance = delta_irradiance
        self.high_res_wavelengths = np.linspace(202, 400, int(np.ceil(198 / 0.01)))
        self.high_res_solar = SolarSpectrum().irradiance(self.high_res_wavelengths)
        self.fwhm_pert_hr = np.linspace(self.fwhm_perturbations[0], self.fwhm_perturbations[-1], 4001)
        self.interp_func = scipy.interpolate.interp2d(self.fwhm_perturbations, self.wavelengths, delta_irradiance)
        self.delta_irradiance = self.interp_func(self.fwhm_pert_hr, self.wavelengths)

    def get_next_fwhm_fsolve(self, center_wavelength, corrected_spectrum_wavelengths,
                             corrected_spectrum, old_fwhm):
        g = make_gaussian(center_wavelength, old_fwhm, self.high_res_wavelengths)
        guess_val = np.sum(g * self.high_res_solar)

        k = np.abs(corrected_spectrum_wavelengths - center_wavelength).argmin()
        wl_idx = np.abs(center_wavelength - self.wavelengths).argmin()
        s_k = self.delta_irradiance[wl_idx, :]

        root_fwhm_pert = find_root(func=self._vec_fun, x0=[self.fwhm_perturbations[0], self.fwhm_perturbations[-1]],
                                   fprime=self._vec_fun_deriv,
                                   args=(self.fwhm_pert_hr, s_k, guess_val, corrected_spectrum[k]))

        # g1 = make_gaussian(center_wavelength, 3.0 + root_fwhm_pert[0], self.high_res_wavelengths)
        # g2 = make_gaussian(center_wavelength, 3.0 + root_fwhm_pert[1], self.high_res_wavelengths)
        # s1 = np.sum(g1 * self.high_res_solar)
        # s2 = np.sum(g2 * self.high_res_solar)
        # delta1 = np.abs(s1 - corrected_spectrum[k])
        # delta2 = np.abs(s2 - corrected_spectrum[k])
        #
        # fwhm_next = 3.0 + root_fwhm_pert[1]
        # delta = delta2
        #
        # if delta1 < delta2:
        #     fwhm_next = 3.0 + root_fwhm_pert[0]
        #     delta = delta1

        fwhm_next = old_fwhm + root_fwhm_pert[np.abs(root_fwhm_pert).argmin()]
        g = make_gaussian(center_wavelength, fwhm_next, self.high_res_wavelengths)
        s = np.sum(g * self.high_res_solar)
        delta = np.abs(s - corrected_spectrum[k])

        rel_spectral_delta = np.abs(delta) / corrected_spectrum[k]
        return fwhm_next, rel_spectral_delta

    def _vec_fun(self, fwhm_perturb_guesses, fwhm_perturbations, perturb_sensitivity, guess_spec_val,
                 corrected_spec_val):
        delta = guess_spec_val - corrected_spec_val
        diff_irrad_normalized = perturb_sensitivity - delta
        idx0 = np.abs(fwhm_perturbations - fwhm_perturb_guesses[0]).argmin()
        idx1 = np.abs(fwhm_perturbations - fwhm_perturb_guesses[1]).argmin()
        return np.array([diff_irrad_normalized[idx0], diff_irrad_normalized[idx1]]).flatten()

    def _vec_fun_deriv(self, fwhm_perturb_guesses, fwhm_perturbations, perturb_sensitivity, guess_spec_val,
                       corrected_spec_val):
        delta = guess_spec_val - corrected_spec_val
        diff_irrad_normalized = perturb_sensitivity - delta
        idx0 = np.abs(fwhm_perturbations - fwhm_perturb_guesses[0]).argmin()
        idx1 = np.abs(fwhm_perturbations - fwhm_perturb_guesses[1]).argmin()

        # use forward difference, unless we are at the end:
        diff0 = 0
        diff1 = 0
        if idx0 < len(fwhm_perturbations) - 1:
            diff0 = (diff_irrad_normalized[idx0 + 1] - diff_irrad_normalized[idx0]) / \
                    (fwhm_perturbations[idx0 + 1] - fwhm_perturbations[idx0])
        else:
            diff0 = (diff_irrad_normalized[idx0] - diff_irrad_normalized[idx0 - 1]) / \
                    (fwhm_perturbations[idx0] - fwhm_perturbations[idx0 - 1])

        if idx1 < len(fwhm_perturbations) - 1:
            diff1 = (diff_irrad_normalized[idx1 + 1] - diff_irrad_normalized[idx1]) / \
                    (fwhm_perturbations[idx1 + 1] - fwhm_perturbations[idx1])
        else:
            diff1 = (diff_irrad_normalized[idx1] - diff_irrad_normalized[idx1 - 1]) / \
                    (fwhm_perturbations[idx1] - fwhm_perturbations[idx1 - 1])

        return np.array([[diff0, 0], [0, diff1]])

    def get_next_fwhm_brute_force(self, center_wavelength, corrected_spectrum_wavelengths,
                                  corrected_spectrum, old_fwhm, step):
        g = make_gaussian(center_wavelength, old_fwhm, self.high_res_wavelengths)
        guess_val = np.sum(g * self.high_res_solar)

        k = np.abs(corrected_spectrum_wavelengths - center_wavelength).argmin()
        wl_idx = np.abs(center_wavelength - self.wavelengths).argmin()
        # s_k = self.interp_func(self.fwhm_perturbations, center_wavelength)
        s_k = self.delta_irradiance[wl_idx, :]
        delta = guess_val - corrected_spectrum[k]

        fprime = np.gradient(s_k)

        # if strictly increasing/decreasing, just invert the function to find the necessary fwhm_delta
        sign = 0
        if np.all(fprime > 0):
            fwhm_delta = np.interp(delta, s_k, self.fwhm_pert_hr)
            sign = int(fwhm_delta / np.abs(fwhm_delta))
        elif np.all(fprime < 0):
            fwhm_delta = np.interp(delta, np.flip(s_k), np.flip(self.fwhm_pert_hr))
            sign = -int(fwhm_delta / np.abs(fwhm_delta))
        else:
            # there's a minimum somewhere, take the second derivative, find that location, then perform the
            # same as above on each side, picking the side with the smaller fwhm adjustment
            fpp = np.gradient(fprime)
            fwhm_delta_minimum_idx = np.abs(fpp).argmin()
            if fprime[0] > fprime[-1]:
                fwhm_delta1 = np.interp(delta, s_k[0:fwhm_delta_minimum_idx],
                                        self.fwhm_pert_hr[0:fwhm_delta_minimum_idx])
                fwhm_delta2 = np.interp(delta, np.flip(s_k[fwhm_delta_minimum_idx + 1:]),
                                        np.flip(self.fwhm_pert_hr[fwhm_delta_minimum_idx + 1:]))
                min_idx = np.abs([fwhm_delta1, fwhm_delta2]).argmin()
                slp = fprime[min_idx * fwhm_delta_minimum_idx + 1]
                sign = int(slp / np.abs(slp))
            else:
                fwhm_delta1 = np.interp(delta, np.flip(s_k[0:fwhm_delta_minimum_idx]),
                                        np.flip(self.fwhm_pert_hr[0:fwhm_delta_minimum_idx]))
                fwhm_delta2 = np.interp(delta, s_k[fwhm_delta_minimum_idx + 1:],
                                        self.fwhm_pert_hr[fwhm_delta_minimum_idx + 1:])
                min_idx = np.abs([fwhm_delta1, fwhm_delta2]).argmin()
                slp = fprime[min_idx * fwhm_delta_minimum_idx + 1]
                sign = -int(slp / np.abs(slp))

        root_fwhm_pert = sign * step
        # just use a small step in the desired direction
        fwhm_next = old_fwhm + root_fwhm_pert
        g1 = make_gaussian(center_wavelength, fwhm_next, self.high_res_wavelengths)
        s1 = np.sum(g1 * self.high_res_solar)
        delta = np.abs(s1 - corrected_spectrum[k])

        rel_spectral_delta = np.abs(delta) / corrected_spectrum[k]
        return fwhm_next, rel_spectral_delta


class AltiusUVSrf(SrfInterface):
    """
        Container class for the wavelength-dependant SRF supplied by ALTIUS
        for the UV channel FPI.
        Provides methods for convolution and chopping off the side lobes.

        Attributes
        ----------
        _data_directory : str
            Path to the ALTIUS SRF data
        _tunable_wavelengths : np.array
            The array of wavelengths that can be exactly tuned to on the instrument, 0.1 nm resolution
        _wavelength_grid : np.array
            High resolution grid of wavelengths each SRF is measured along, 0.01 nm resolution
        _fpi_coef : np.array
            Tunable parameters for the Fabry-Perot Interferometer at each tunable wavelength
        _fpi_srf : np.array
            The +srf at each tunable wavelength overt the wavelength grid.
            An array len(_tunable_wavelengths) x len(_wavelength_grid)
        _optics_srf : np.array
            The transfer function as a function of wavelength due to the optics
        _detector_q : np.array
            The transfer function as a function of wavelength of the detector quantum efficiency
        LINEAR_INTERP : str
            Constant for specifying linear interpolation
        LOG_INTERP : str
            Constant for specifying log interpolation
        fwhm_method_using_area : bool
            If True, the FWHM of the actual SRF is computed by computing the area under the central lobe.
            If False, the FWHM is computed by determining the width of the half-max points of central lobe.

        Methods
        -------
        get_srf(center_wavelength_nm, new_wavelengths=None, interp_method='linear', normalize=True):
            Gets the SRF at the requested center wavelength.

        get_gaussian(center_wavelength_nm, new_wavelengths):
            Computes the Gaussian that closely matches the FWHM of the SRF at the center wavelength

        convolve_full_srf(spectrum_wavelengths: np.ndarray,
                          spectrum: np.ndarray,
                          convolution_wavelengths: np.ndarray):
            Convolves the full SRF data at the convolution wavelengths with the input spectrum

        convolve_corrected_srf(spectrum_wavelengths: np.ndarray,
                               spectrum: np.ndarray,
                               convolution_wavelengths: np.ndarray,
                               convolved_spectrum_full_srf: xr.DataArray)
            Convolves the corrected SRF at the convolution wavelengths with the input spectrum

        spectrum_central_lobe_sensitivity(spectrum_wavelengths: np.ndarray,
                                          spectrum: np.ndarray) -> SrfSensitivity:
            Computes the spectrum sensitivity from 250 to 350 nm for perturbations of
            -1.0 to 3.0 nm in the FWHM of the Gaussian-approximated SRF.

        modify_center_lobe_width(center_lobe: np.ndarray,
                                 center_lobe_wavelengths: np.ndarray,
                                 scale: np.ndarray,
                                 center_wavelength: float = None):
            Given the center lobe (side lobe removed), modifies the center lobe width
            by scale.  Essentially the function squished the wavelengths around the center
            wavelength to make the center peak "wider"

        get_gaussian_contribution(spectrum_wavelengths, spectrum, convolution_wavelengths):
            Computes the contribution across the convolution_wavelengths due to a Gaussian
            of equivalent FWHM to the SRF at each wavelength.

        compute_srf_approximate_fwhm(center_wavelength_nm, new_wavelengths):

        center_lobe_adjusted_fwhm(center_wavelength_nm, eval_wavelengths, fwhm):
            Partitions the SRF at the center wavelength in side band and central lobe, and returns
            the center lobe adjusted to have the requested FWHM

        partition_srf(center_wavelength_nm, new_wavelengths):
            Selects the SRF at the requested wavelength and divides it into the center lobe
            and the side bands according to _get_center_lobe_index_span
        """

    # LINEAR_INTERP = 'linear'
    # LOG_INTERP = 'log'

    def __init__(self, data_directory, fwhm_method_using_area=False):
        self._data_directory = data_directory
        self._tunable_wavlengths, self._wavelength_grid, self._fpi_coef, self._fpi_srf = self._load_fpi_srf()
        self._optics_srf = self._load_optics_srf(self._wavelength_grid).flatten()
        self._detector_qe = self._load_detector_qe(self._wavelength_grid).flatten()
        self.fwhm_method_using_area = fwhm_method_using_area

    def tunable_wavelengths(self):
        return self._tunable_wavlengths

    def get_base_wavelength_measurement_grid(self):
        return self._wavelength_grid

    def get_wavelength_grid(self, resolution=None):
        default_grid = np.linspace(202, 400, int(198 / RESOLUTION) + 1)
        if resolution is None:
            return default_grid

        return np.arange(202, 400, step=resolution)

    def fpi_coefficients(self):
        return self._fpi_coef

    def all_fpi_srf(self):
        return self._fpi_srf

    def optics_srf(self):
        return self._optics_srf

    def detector_qe(self):
        return self._detector_qe

    def _load_fpi_srf(self) -> (np.ndarray, np.ndarray, np.ndarray, np.ndarray):
        csv.field_size_limit(int(sys.maxsize / 10000000000))
        data = []
        fname = os.path.join(self._data_directory, 'srf_uv.txt')
        with open(fname) as f:
            rd = csv.reader(f)
            for i, row in enumerate(rd):
                line = []
                x = row[0].split()
                for v in x:
                    line.append(float(v))
                data.append(line)

        data = np.array(data)

        tunable_wavelengths = data[1:, 0]
        fpi_coef = data[1:, 1:5]
        wavelengths = data[0, 5:]
        srf = data[1:, 5:]

        return tunable_wavelengths.flatten(), wavelengths.flatten(), fpi_coef, srf

    def _load_optics_srf(self, wavelengths) -> np.ndarray:
        fname = os.path.join(self._data_directory, 'uv_optics_srf_11.txt')
        optics = np.genfromtxt(fname, dtype=np.float, skip_header=1)

        # interpolate optics WL (column 0) BL_BOL (column 1) onto +srf grid (column 0)
        optics_filter = np.interp(wavelengths.flatten(), optics[:, 0].flatten(), optics[:, 1].flatten(), right=0,
                                  left=0)
        optics_filter = optics_filter[:, np.newaxis]
        return optics_filter

    def _load_detector_qe(self, wavelengths) -> np.ndarray:
        fname = os.path.join(self._data_directory, 'SRF_uv_data.mat')
        srf_data_dictionary = scipy.io.loadmat(fname)
        data = srf_data_dictionary["SRF"]
        wl = srf_data_dictionary["SRF_WL"].flatten() * 1e9

        detector_srf = data[0][5][:, 0, 0].flatten()
        qe = np.interp(wavelengths.flatten(), wl, detector_srf)
        qe = qe[:, np.newaxis]
        return qe

    def get_srf(self, center_wavelength_nm, new_wavelengths=None,
                interp_method=SrfInterface.LINEAR_INTERP, normalize=True) -> Tuple[xr.DataArray, float]:
        """
        Returns the SRF closest to the specified wavelength, and the actual wavelength

        Parameters
        ----------
        center_wavelength_nm : float
            Center wavelength to obtain the SRF from the tunable wavelengths

        new_wavelengths : np.ndarray, Optional
            The grid of wavelengths, in nm, to evaluate/interpolate the SRF over.  If not
            specified, will return the SRF at the original measured wavelengths in wavelength_grid

        interp_method : str, Optional
            'linear' or 'log'.  Interpolation method for evaluating the SRF over new_wavelengths

        normalize : bool, Optional
            If True, normalizes the integral of the SRF to be 1.0.
            If False, the values of the SRF are adjusted so that the integral matches the original
            data's integral

        Returns
        -------
        +srf : xr.DataArray
            The SRF with tuned wavelength near to center_wavelength_nm.  x-coordinate 'wavelength' set
            by the new_wavelengths parameter
        actual_wavelength : float
            The nearest tunable wavelength to the input center_wavelength_nm
        """
        tunable_idx = np.abs(self._tunable_wavlengths - center_wavelength_nm).argmin()
        srf = self._fpi_srf[tunable_idx, :] * self._detector_qe * self._optics_srf

        orig_sum = np.sum(srf)

        if new_wavelengths is not None:
            if interp_method is self.LINEAR_INTERP:
                srf = np.interp(new_wavelengths.flatten(), self._wavelength_grid.flatten(), srf.flatten(), left=0,
                                right=0)
            else:
                log_srf = np.log(srf.flatten())
                log_srf = np.interp(new_wavelengths.flatten(), self._wavelength_grid.flatten(), log_srf, left=0,
                                    right=0)
                srf = np.exp(log_srf)

            srf = srf * orig_sum / np.sum(srf)  # make sure it scales to the original SRF data
            srf = xr.DataArray(srf, dims={'wavelength'}, coords={'wavelength': new_wavelengths})
        else:
            srf = xr.DataArray(srf, dims={'wavelength'}, coords={'wavelength': self._wavelength_grid})

        if normalize:
            srf /= np.sum(srf)

        return srf, self._tunable_wavlengths[tunable_idx]

    def get_gaussian(self, center_wavelength_nm, new_wavelengths) -> Tuple[np.ndarray, float, float]:
        """
        Gets the Gaussian PSF across new_wavelengths centered at center_wavelength_nm.  The Gaussian
        has the same FWHM as the SRF at the center wavelength, but not the same integral.  Instead,
        it is normalized to 1.0.

        Parameters
        ----------
        center_wavelength_nm : float
            Center wavelength to approximate the SRF with a Gaussian

        new_wavelengths : np.ndarray
            The grid of wavelengths, in nm, to evaluate the Gaussian over

        Returns
        -------
        blur : np.ndarray
            The Gaussian PSF with maximum at center_wavelength_nm, evaluated on new_wavelengths
        actual_wl : float
            The nearest tunable wavelength to center_wavelength_nm
        fwhm : float
            The FWHM in nm determined for the actual SRF at center_wavelength_nm
        """
        fwhm, actual_wl = self.compute_srf_approximate_fwhm(center_wavelength_nm, new_wavelengths)
        sigma = fwhm / (2 * np.sqrt(2 * np.log(2)))
        blur = 1 / np.sqrt(2 * np.pi * sigma ** 2) * np.exp(-0.5 * (actual_wl - new_wavelengths) ** 2 / sigma ** 2)
        blur /= np.sum(blur)

        return blur, actual_wl, fwhm

    def convolve_full_srf(self, spectrum_wavelengths: np.ndarray,
                          spectrum: np.ndarray,
                          convolution_wavelengths: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Convolves the spectrum with the SRF at each wavelength.

        Parameters
        ----------
        spectrum_wavelengths : np.ndarray
            Wavelengths in nm of the input spectrum

        spectrum : np.ndarray
            Input high-resolution spectrum to be convolved with the corrected SRF

        convolution_wavelengths : np.ndarray
            Wavelengths (in nm) to evaluate the resulting convolution at

        Returns
        -------
        convolved_side_lobes_removed : np.ndarray
            The resulting spectrum convolved with the corrected SRFs with side lobes removed

        central_lobe_portion : np.ndarray
            Relative contribution of the central lobe portion to each convolution wavelength compared
            to the total SRF

        central_lobe_fwhm : np.ndarray
            FWHM in nm computed for the SRF at each convolution wavelength
        """
        # TODO assert spectrum_wavelengths has sufficient padding outside of convolution_wavelengths
        convolved_spectrum = self.convolve_spectrum_with_srf(convolution_wavelengths, spectrum_wavelengths, spectrum)

        central_lobe_fwhm, central_lobe_portion = self._compute_central_lobe_characteristics(convolution_wavelengths,
                                                                                             spectrum,
                                                                                             spectrum_wavelengths)

        return convolved_spectrum, central_lobe_portion, central_lobe_fwhm

    def convolve_spectrum_with_srf(self, convolution_wavelengths: np.ndarray,
                                   spectrum_wavelengths: np.ndarray,
                                   spectrum: np.ndarray):
        """
            Convolves the spectrum with the SRF at each wavelength.

            Parameters
            ----------
            spectrum_wavelengths : np.ndarray
                Wavelengths in nm of the input spectrum

            spectrum : np.ndarray
                Input high-resolution spectrum to be convolved with the corrected SRF

            convolution_wavelengths : Union[float, np.ndarray]
                Wavelengths (in nm) to evaluate the resulting convolution at

            Returns
            -------
            convolved_spectrum : np.ndarray
                The resulting spectrum convolved with SRFs at each convolution wavelength
        """
        convolved_spectrum = np.zeros_like(convolution_wavelengths)
        for k, w in enumerate(convolution_wavelengths):
            srf_w, _ = self.get_srf(w, spectrum_wavelengths, normalize=False)
            srf_contribution_spectrum = srf_w.values * spectrum
            srf_contribution = np.sum(srf_contribution_spectrum)
            convolved_spectrum[k] = srf_contribution
        return convolved_spectrum

    def _compute_central_lobe_characteristics(self, convolution_wavelengths, spectrum, spectrum_wavelengths):
        central_lobe_portion = 0 * convolution_wavelengths
        central_lobe_fwhm = 0 * convolution_wavelengths
        for k, w in enumerate(convolution_wavelengths):
            srf_w, _ = self.get_srf(w, spectrum_wavelengths, normalize=False)
            srf_contribution_spectrum = srf_w * spectrum
            srf_contribution = np.sum(srf_contribution_spectrum)
            fwhm, _ = self.compute_srf_approximate_fwhm(w, spectrum_wavelengths)
            central_lobe_fwhm[k] = fwhm
            idx_span = self._get_center_lobe_index_span(w, spectrum_wavelengths)
            central_lobe_area = np.sum(srf_contribution_spectrum[idx_span])
            p = central_lobe_area / srf_contribution
            central_lobe_portion[k] = p
        return central_lobe_fwhm, central_lobe_portion

    def convolve_corrected_srf(self, spectrum_wavelengths: np.ndarray,
                               spectrum: np.ndarray,
                               convolution_wavelengths: np.ndarray,
                               convolved_spectrum_full_srf: xr.DataArray):
        """
        Convolves the spectrum with the corrected SRF at each wavelength.  The corrected SRF at a particular
        wavelength is the SRF with the side lobes removed, and re-normalized to be the original integrated

        Parameters
        ----------
        spectrum_wavelengths : np.ndarray
            Wavelengths in nm of the input spectrum

        spectrum : np.ndarray
            Input high-resolution spectrum to be convolved with the corrected SRF

        convolution_wavelengths : np.ndarray
            Wavelengths (in nm) to evaluate the resulting convolution at

        convolved_spectrum_full_srf : xr.DataArray
            The previously convolved spectrum with the full, uncorrected SRF

        Returns
        -------
        convolved_side_lobes_removed : np.ndarray
            The resulting spectrum convolved with the corrected SRFs with side lobes removed

        central_lobe_portion : np.ndarray
            Relative contribution of the central lobe portion to each convolution wavelength compared
            to the total SRF

        central_lobe_fwhm : np.ndarray
            FWHM in nm computed for the SRF at each convolution wavelength

        srf_central_lobe_sum : np.ndarray
            The integral underneath the central lobe for each convolution wavelength for the convolved
            spectrum.

        srf_side_lobe_sum : np.ndarray
            The integral outside the central lobe for each convolution wavelength for the
            convolved spectrum
        """
        # TODO assert spectrum_wavelengths has sufficient padding outside of convolution_wavelengths
        convolved_spectrum = np.zeros_like(convolution_wavelengths)
        central_lobe_portion = np.zeros_like(convolution_wavelengths)
        central_lobe_fwhm = np.zeros_like(convolution_wavelengths)
        convolved_side_lobes_removed = np.zeros_like(convolution_wavelengths)
        srf_central_lobe_sum = np.zeros((len(convolution_wavelengths), 2))
        srf_side_lobe_sum = np.copy(srf_central_lobe_sum)

        srf_full_contribution = convolved_spectrum_full_srf.interp(wavelength=convolution_wavelengths).values

        for k, w in enumerate(convolution_wavelengths):
            srf_w, _ = self.get_srf(w, spectrum_wavelengths, normalize=False)
            center_lobe, side_band_only, central_lobe_sum = self.partition_srf(w, spectrum_wavelengths)
            side_band_sum = np.sum(side_band_only)
            srf_contribution = srf_full_contribution[k]
            side_contribution = np.sum(side_band_only * spectrum)
            convolved_side_lobes_removed[k] = (srf_contribution - side_contribution) / central_lobe_sum
            convolved_spectrum[k] = srf_contribution
            fwhm, _ = self.compute_srf_approximate_fwhm(w, self.get_base_wavelength_measurement_grid())
            central_lobe_fwhm[k] = fwhm
            p = central_lobe_sum / srf_contribution
            central_lobe_portion[k] = p
            srf_central_lobe_sum[k] = (central_lobe_sum, srf_contribution - side_contribution)
            srf_side_lobe_sum[k] = (side_band_sum, side_contribution)

        return convolved_side_lobes_removed, central_lobe_portion, central_lobe_fwhm, srf_central_lobe_sum, srf_side_lobe_sum

    def spectrum_central_lobe_sensitivity(self, spectrum_wavelengths: np.ndarray,
                                          spectrum: np.ndarray) -> SrfSensitivity:
        """
        Computes the spectrum sensitivity from 250 to 350 nm for perturbations of
        -1.0 to 3.0 nm in the FWHM of the Gaussian-approximated SRF.
        """
        # add a +/-0.01 nm, +/0.02, ..., +/-2.0 nm perturbation to the central lobe width, get the sensitivity for each perturbation
        res = spectrum_wavelengths[1] - spectrum_wavelengths[0]

        convolution_wavelengths = np.linspace(250, 350, 1001)  # full resolution here

        perturb = np.linspace(-1.0, 3.0, 41)  # 0.1 nm grid
        sensitivity = np.zeros((len(convolution_wavelengths), len(perturb)))

        # num_cores = multiprocessing.cpu_count()
        # sensitivity = Parallel(n_jobs=num_cores) \
        #         (delayed(self._compute_sensitivity_at_center_wavelength(perturb, spectrum, spectrum_wavelengths, w)
        #                  for k, w in enumerate(convolution_wavelengths)))
        pb = progress_bar.ProgressBar(0, len(convolution_wavelengths))
        print('Computing sensitivity')
        for k, w in enumerate(convolution_wavelengths):
            s_k = self._compute_sensitivity_at_center_wavelength(perturb, spectrum, spectrum_wavelengths, w)
            sensitivity[k] = s_k
            pb.update(k)
        print('Finished computing sensitivity')
        # sensitivity_grid = RectBivariateSpline(convolution_wavelengths, perturb, sensitivity)
        return SrfSensitivity(perturb, convolution_wavelengths, sensitivity)

    def _compute_sensitivity_at_center_wavelength(self, fwhm_perturbations, spectrum_values, spectrum_wavelengths,
                                                  center_wavelength):
        """
        Computes the sensitivity of an equivalent Gaussian at the center wavelength.

        Parameters
        ----------
        fwhm_perturbations : np.ndarray
            Array of changes in nm to the FWHM to evaluate the sensitivity

        spectrum_wavelengths : np.ndarray
            Array of wavelengths in nm of the spectrum

        spectrum_values : np.ndarray
            Spectrum evaluated at spectrum_wavelengths

        center_wavelength : float
            Wavelength to compute the sensitivity at.  Used to select the SRF
        """
        # center_lobe, side_band_only, central_lobe_sum = self.partition_srf(center_wavelength, spectrum_wavelengths)
        center_lobe, _, _ = self.get_gaussian(center_wavelength, spectrum_wavelengths)
        fwhm, _ = self.compute_srf_approximate_fwhm(center_wavelength, spectrum_wavelengths)
        scale = (fwhm + fwhm_perturbations) / fwhm
        wider_central_lobe = self.modify_center_lobe_width(center_lobe, spectrum_wavelengths, scale, center_wavelength)
        s_k = (np.sum(wider_central_lobe * spectrum_values, axis=1) - np.sum(center_lobe * spectrum_values))
        s_k = np.nan_to_num(s_k, nan=0.0, posinf=0.0, neginf=0.0)
        return s_k

    def modify_center_lobe_width(self, center_lobe: np.ndarray,
                                 center_lobe_wavelengths: np.ndarray,
                                 scale: np.ndarray,
                                 center_wavelength: float = None):
        """
        Given the center lobe (side lobe removed), modifies the center lobe width
        by scale.  Essentially the function squished the wavelengths around the center
        wavelength to make the center peak "wider"

        Parameters
        ----------
        center_lobe : np.ndarray
            The SRF at some center wavelength with the side lobes removed

        center_lobe_wavelengths : np.ndarray
            Wavelengths in nm of the array above.  Length n

        scale : np.ndarray
            Scale at which to modify the FWHM of the center lobe.  Arbitrarily this
            is a range of numbers for which we'll scale the center lobe.  Length m

        center_wavelength : float, Optional
            The center wavelength of the passed-in center_lobe.  If not passed in,
            it will assume the max of the center lobe is the at the center wavelength.
            In most cases this is appropriate, but some of the SRFs are poorly behaved and
            their peaks are offset from the actual center wavelength.  Length n

        Returns
        -------
        wide_lobe_all_scales : np.ndarray
            The center lobe with a FWHM altered by the values set in scale.  Each array
            is the set of center_lobe_wavelengths with a lobe scaled by a single value in
            the array scale.  m x n array
        """
        central_lobe_sum = np.sum(center_lobe)
        center_lobe_wavelengths = center_lobe_wavelengths[np.newaxis, ...]
        scale = scale[..., np.newaxis]
        if center_wavelength is None:
            idx = np.argmax(center_lobe)
            center_wavelength = center_lobe_wavelengths.flatten()[idx]
        squished_wavelengths = scale * (center_lobe_wavelengths - center_wavelength) + center_wavelength

        wide_lobe_all_scales = 0 * squished_wavelengths
        # interpolate back onto the actual grid:
        for k, sw in enumerate(squished_wavelengths):
            wider_central_lobe = np.interp(center_lobe_wavelengths, sw, center_lobe)
            # wider_central_lobe = wider_central_lobe * central_lobe_sum / np.sum(wider_central_lobe)
            wider_central_lobe /= np.sum(wider_central_lobe)
            wide_lobe_all_scales[k] = wider_central_lobe

        return wide_lobe_all_scales

    def _get_approximate_central_lobe_window(self, res):
        """
        Simple placeholder for choosing the central lobe window : 5.0 nm wide
        """
        return int(5.0 / res)

    def get_gaussian_contribution(self, spectrum_wavelengths, spectrum, convolution_wavelengths):
        """
        Computes the contribution across the convolution_wavelengths due to a Gaussian
        of equivalent FWHM to the SRF at each wavelength.

        Parameters
        ----------
        spectrum_wavelengths : np.ndarray
            Wavelengths in nm of the input spectrum

        spectrum : np.ndarray
            Spectrum to be convolved with Gaussians

        convolution_wavelengths : np.ndarray
            Evaluation wavelengths for the convolution, in nm

        Returns
        -------
        central_lobe_portion_gs : np.ndarray
            Convolution of the input spectrum with Gaussians evaluated at convolution_wavelengths.
            Each Gaussian is computed to have the same FWHM at the convolution wavelength as
            the actual SRF.
        """
        central_lobe_portion_gs = np.zeros(len(convolution_wavelengths))
        res = spectrum_wavelengths[1] - spectrum_wavelengths[0]
        for k, w in enumerate(convolution_wavelengths):
            g, _, _ = self.get_gaussian(w, spectrum_wavelengths)
            gauss_contrib = g * spectrum
            idx_span = self._get_center_lobe_index_span(w, spectrum_wavelengths)
            p_g = np.sum(gauss_contrib[idx_span]) / np.sum(gauss_contrib)
            central_lobe_portion_gs[k] = p_g

        return central_lobe_portion_gs

    def compute_srf_approximate_fwhm(self, center_wavelength_nm, new_wavelengths):
        srf, actual_wl = self.get_srf(center_wavelength_nm, new_wavelengths, normalize=False)
        fwhm_area_method, fwhm_width_method = self._fwhm_of_srf(actual_wl, new_wavelengths, srf.values)
        fwhm = fwhm_width_method
        if self.fwhm_method_using_area:
            fwhm = fwhm_area_method
        return fwhm, actual_wl

    def _fwhm_of_srf(self, center_wavelength_nm, srf_wavelengths, srf):
        """
        Computes the FWHM of the actual SRF data at a requested center wavelength.
        The FWHM is computed in two ways - by evaluating the integral of the central lobe,
        and by interpolating the half-max points.  The reasoning for this is because the
        SRF central lobe is not an ideal Gaussian, and conventional calculations of
        the FWHM assuming a Gaussian shape yield inconsistent results.
        The choice of calculated FWHM is set in the class variable self.fwhm_method_using_area

        The "area method" involves determining the area under a Gaussian by
        looking at the CDF of the Gaussian distribution at :math:`-\sqrt{2 ln2}`.
        The FWHM of a Gaussian is :math:`2\sqrt{2 ln2}\sigma, so therefore the
        area within the FWHM range of a standard normal is:

        .. math::
                    area = 1.0 - 2\,CDF(-\sqrt{2\,ln2})

        Where the CDF is the cumulative distribution function of a standard normal.
        The cumulative sum of the SRF data is determined at each wavelength across the central
        lobe, and the area is interpolated to the FWHM area above to find the wavelength
        range of the SRF that gives the approximated FWHM.

        TODO the central lobe SRF should be normalized to 1.0 when computing this way

        Returns
        -------
        fwhm_area_method : float

        fwhm_width_method : float

        """
        res = srf_wavelengths[1] - srf_wavelengths[0]
        area_in_fwhm = 1 - 2 * norm_dist(0, 1).cdf(-np.sqrt(2 * np.log(2)))
        central_lobe_window = self._get_approximate_central_lobe_window(res)
        # max_srf_idx = np.abs(eval_wavelengths - center_wavelength_nm).argmin()
        # changed to finding the max index, since many SRF's aren't centered on the tunable wavelength
        max_srf_idx = np.argmax(srf)
        window_srf_left = srf[max_srf_idx - central_lobe_window:max_srf_idx + 1]
        delta_wl_left = np.abs(
            srf_wavelengths[max_srf_idx - central_lobe_window:max_srf_idx + 1] - center_wavelength_nm)
        half_peak_val = 0.5 * srf[max_srf_idx]
        hwhm_left = np.interp(half_peak_val, window_srf_left, delta_wl_left)
        window_srf_right = srf[max_srf_idx:max_srf_idx + central_lobe_window + 1]
        delta_wl_right = np.abs(
            srf_wavelengths[max_srf_idx:max_srf_idx + central_lobe_window + 1] - center_wavelength_nm)
        wsr = np.copy(window_srf_right)
        wsr[0] *= 0.5
        wsl = np.copy(window_srf_left)
        wsl[-1] *= 0.5
        area = (np.cumsum(np.flip(wsl)) + np.cumsum(wsr)) / (np.sum(wsr) + np.sum(wsl))
        fwhm_area_method = 2 * np.interp(area_in_fwhm, area, delta_wl_right)
        hwhm_right = np.interp(half_peak_val, np.flipud(window_srf_right), np.flipud(delta_wl_right))
        fwhm_width_method = hwhm_left + hwhm_right
        return fwhm_area_method, fwhm_width_method

    def center_lobe_adjusted_fwhm(self, center_wavelength_nm, eval_wavelengths, fwhm):
        """
        Partitions the SRF at the center wavelength in side band and central lobe, and returns
        the center lobe with the FWHM requested

        Parameters
        ----------
        center_wavelength_nm : float
            Requested center wavelength in nanometers

        eval_wavelengths : np.ndarray
            The gridded set of wavelengths to return the adjusted center lobe on

        fwhm : float
            The requested FWHM in nanometers - the center lobe will be adjusted to this width

        Returns
        -------
        modified_center_lobe : np.ndarray
            SRF centered at center_wavelength_nm, evaluated at the eval_wavelengths and modified to
            have a FWHM of fwhm.  The side lobes have been removed, and the integral of this SRF is
            normalized to be the same as the original data
        """
        center_lobe, _, _ = self.partition_srf(center_wavelength_nm, eval_wavelengths)
        actual_fwhm, _ = self.compute_srf_approximate_fwhm(center_wavelength_nm, eval_wavelengths)
        scale = fwhm / actual_fwhm
        modified_center_lobe = self.modify_center_lobe_width(center_lobe, eval_wavelengths, scale)
        return modified_center_lobe

    def partition_srf(self, center_wavelength_nm, new_wavelengths) -> (np.ndarray, np.ndarray, float):
        """
        This function selects the SRF at the requested wavelength and divides it into the center lobe
        and the side bands according to _get_center_lobe_index_span

        Parameters
        ----------
        center_wavelength_nm : float
            Center wavelength of the SRF requested in nm

        new_wavelengths : np.ndarray
            Gridded wavelengths to evaluate the SRF at, in nm

        Returns
        -------
        center_lobe : np.ndarray
            SRF at the center wavelength with the side bands removed.  Evaluated at
            the gridded wavelengths, and re-normalized to the original integral of the
            SRF.

        side_bands : np.ndarray
            The side bands that were removed from the overall SRF.  The center lobe is set
            to zero for this array.

        center_lobe_sum : float
            The integral of the center lobe
        """
        srf, _ = self.get_srf(center_wavelength_nm, new_wavelengths, normalize=False)
        center_lobe_index = self._get_center_lobe_index_span(center_wavelength_nm, new_wavelengths)
        center_lobe_sum = np.sum(srf.values[center_lobe_index])

        side_bands = np.copy(srf)
        side_bands[center_lobe_index] = 0

        center_lobe = np.copy(srf)
        center_lobe[~center_lobe_index] = 0

        return center_lobe, side_bands, center_lobe_sum

    def _get_center_lobe_index_span(self, center_wavelength_nm, new_wavelengths):
        """
        Gets the index span of the center lobe for the SRF at a center wavelength.
        The chosen index span currently goes from -fwhm to +fwhm relative to
        the center wavelength.

        Parameters
        ----------
        center_wavelength_nm : float
            Center wavelength of the SRF requested in nm

        new_wavelengths : np.ndarray
            Gridded wavelengths to evaluate the SRF at, in nm

        Returns
        -------
        idx : np.ndarray
            Mask array of same length as new_wavelengths specifying the center lobe
        """
        fwhm, actual_wl = self.compute_srf_approximate_fwhm(center_wavelength_nm, new_wavelengths)
        idx = (new_wavelengths >= center_wavelength_nm - fwhm) & (new_wavelengths <= center_wavelength_nm + fwhm)
        return idx


class FwhmPerturbations:

    def __init__(self, srf: AltiusUVSrf, solar_distance_AU=1.0, hr_wavelengths: np.ndarray = None, scaling_factor=None):
        # self._fwhm_perturbs = np.linspace(-2.0, 5.0, 56)  # 0.1 nm spacing
        # res = np.min(SolarSpectrum().resolution([202, 400]))
        self._fwhm_perturbs = np.linspace(-0.1, 0.1, 21)  # 0.01 nm spacing
        if hr_wavelengths is not None:
            self._high_res_wavelengths = hr_wavelengths
        else:
            self._high_res_wavelengths = srf.get_wavelength_grid()

        self._high_res_scaled_solar_spectrum = SolarSpectrum().irradiance(self._high_res_wavelengths,
                                                                          solardistance=solar_distance_AU)

        if scaling_factor is not None:
            self._high_res_scaled_solar_spectrum *= scaling_factor
        self._srf = srf

    def step_fwhm_assume_constant_srf_central_lobe_shape(self, old_fwhm, center_wavelength,
                                                         corrected_measured_val,
                                                         previous_delta_spectrum=None) -> (float, float):

        srf_scaled_to_old_fwhm = self._srf.center_lobe_adjusted_fwhm(center_wavelength,
                                                                     self._high_res_wavelengths,
                                                                     old_fwhm)

        # old_gaussian = make_gaussian(center_wavelength, old_fwhm, self._high_res_wavelengths)
        return self._compute_next_fwhm_using_center_lobe_shape(center_wavelength,
                                                               corrected_measured_val, old_fwhm,
                                                               srf_scaled_to_old_fwhm,
                                                               previous_delta_spectrum)

    def step_fwhm_assume_gaussian(self, old_fwhm, center_wavelength,
                                  corrected_measured_val,
                                  previous_delta_spectrum=None):
        old_gaussian = make_gaussian(center_wavelength, old_fwhm, self._high_res_wavelengths)
        return self._compute_next_fwhm_using_center_lobe_shape(center_wavelength,
                                                               corrected_measured_val, old_fwhm,
                                                               old_gaussian,
                                                               previous_delta_spectrum)

    def _compute_next_fwhm_using_center_lobe_shape(self, center_wavelength,
                                                   corrected_measured_val, old_fwhm,
                                                   center_lobe_scaled_to_old_fwhm,
                                                   previous_delta_spectrum=None):
        scale = (old_fwhm + self._fwhm_perturbs) / old_fwhm
        wider_central_lobe = self.modify_center_lobe_width(center_lobe_scaled_to_old_fwhm,
                                                           self._high_res_wavelengths, scale,
                                                           center_wavelength)
        old_spectrum_value = np.sum(center_lobe_scaled_to_old_fwhm * self._high_res_scaled_solar_spectrum)
        altered_fwhm_spectrum_values = np.sum(wider_central_lobe * self._high_res_scaled_solar_spectrum, axis=1)
        fwhm_sensitivity = altered_fwhm_spectrum_values - old_spectrum_value
        fwhm_sensitivity = np.nan_to_num(fwhm_sensitivity, nan=0.0, posinf=0.0, neginf=0.0)

        hr_fwhm_pert = self._fwhm_perturbs
        hr_fwhm_sensitivity = fwhm_sensitivity
        next_fwhm, delta_spectrum = self._compute_next_fwhm(center_wavelength, corrected_measured_val, hr_fwhm_pert,
                                                            hr_fwhm_sensitivity, old_fwhm, old_spectrum_value,
                                                            previous_delta_spectrum)
        return next_fwhm, delta_spectrum

    def _compute_next_fwhm(self, center_wavelength, corrected_measured_val,
                           hr_fwhm_pert, hr_fwhm_sensitivity, old_fwhm, old_spectrum_value,
                           previous_delta=None):
        current_meas_delta = corrected_measured_val - old_spectrum_value
        min_idx = np.argsort(np.abs(hr_fwhm_sensitivity - current_meas_delta))
        fwhm_perturbs_giving_smallest_delta = hr_fwhm_pert[min_idx[:10]]
        smallest_delta_perturbations = hr_fwhm_sensitivity[min_idx[:10]]
        # choose the two smallest perturb deltas, then choose the smaller fwhm delta (don't want to take big steps)
        sort_idx = np.argsort(smallest_delta_perturbations)
        delta_fwhm = fwhm_perturbs_giving_smallest_delta[
            np.argmin(np.abs(fwhm_perturbs_giving_smallest_delta[sort_idx[:2]]))]
        # delta_fwhm = fwhm_perturbs_giving_smallest_delta[np.abs(smallest_delta_perturbations).argmin()]
        new_fwhm = old_fwhm + delta_fwhm
        new_gaussian = make_gaussian(center_wavelength, new_fwhm, self._high_res_wavelengths)
        delta_spectrum = corrected_measured_val - np.sum(new_gaussian * self._high_res_scaled_solar_spectrum)

        return new_fwhm, delta_spectrum

    def _check_for_inflection(self, center_wavelength, corrected_measured_val, current_meas_delta, delta_fwhm,
                              delta_spectrum, hr_fwhm_pert, hr_fwhm_sensitivity, new_fwhm, old_fwhm):
        fp = np.gradient(hr_fwhm_sensitivity)
        changed_delta_spectrum = delta_spectrum
        changed_new_fwhm = new_fwhm
        if ~np.all(fp > 0) and ~np.all(fp < 0):
            inflection_guess_indices = np.argsort(np.abs(fp))
            fwhm_perturbs_at_inflection = hr_fwhm_pert[inflection_guess_indices[:10]]
            delta_perturbs = hr_fwhm_sensitivity[inflection_guess_indices[:10]]
            # choose the one closest to the actual current delta:
            try_delta_fwhm = fwhm_perturbs_at_inflection[np.abs(delta_perturbs - current_meas_delta).argmin()]
            try_new_fwhm = old_fwhm + try_delta_fwhm
            new_gaussian = make_gaussian(center_wavelength, try_new_fwhm, self._high_res_wavelengths)
            try_delta_spectrum = corrected_measured_val - np.sum(new_gaussian * self._high_res_scaled_solar_spectrum)

            # take a smaller step if possible
            if np.abs(try_delta_fwhm) < np.abs(delta_fwhm):
                changed_new_fwhm = try_new_fwhm
                changed_delta_spectrum = try_delta_spectrum
        return changed_delta_spectrum, changed_new_fwhm

    @staticmethod
    def modify_center_lobe_width(center_lobe: np.ndarray, spectrum_wavelengths: np.ndarray, scale: np.ndarray,
                                 center_wavelength: float = None):
        # squish the wavelengths around this center wavelength to make the center peak "wider"
        central_lobe_sum = np.sum(center_lobe)
        spectrum_wavelengths = spectrum_wavelengths[np.newaxis, ...]
        scale = scale[..., np.newaxis]
        if center_wavelength is None:
            idx = np.argmax(center_lobe)
            center_wavelength = spectrum_wavelengths.flatten()[idx]
        squished_wavelengths = scale * (spectrum_wavelengths - center_wavelength) + center_wavelength

        wide_lobe_all_scales = 0 * squished_wavelengths
        # interpolate back onto the actual grid:
        for k, sw in enumerate(squished_wavelengths):
            wider_central_lobe = np.interp(spectrum_wavelengths.flatten(), sw.flatten(), center_lobe.flatten())
            wider_central_lobe = wider_central_lobe * central_lobe_sum / np.sum(wider_central_lobe)
            wide_lobe_all_scales[k] = wider_central_lobe

        return wide_lobe_all_scales
