import os

import numpy as np
import progressbar
import xarray as xr
from sasktran.solarspectrum import SolarSpectrum
# import logging
from scipy.optimize import least_squares

from altius_l1pp_inflight.srf.blurred_spectra import SAO2010
from altius_l1pp_inflight.srf.uv_srf import AltiusUVSrf, FwhmPerturbations

script_dir = os.path.dirname(__file__)
# rel_path = '../../../altius_l1pp_sphinx_documentation/sphinx/source/special_studies/2020-06-15_wavelength_registration/sphinx/figures/uniform_run/'
# SAVE_DIR = os.path.join(script_dir, rel_path)
FWHM_CONVERT = 2 * np.sqrt(2 * np.log(2))


def fwhm_2_sigma(fwhm):
    return fwhm / FWHM_CONVERT


def sigma_2_fwhm(sigma):
    return FWHM_CONVERT * sigma


class SimulatedUVSolarSpectrum:
    """Blurs the high-res SAO2010 solar spectrum with a wavelength-dependent SRF, and
    adds in the side lobe correction assuming perfect knowledge of the SRF"""

    FULL_SRF = 'full'
    SIDE_LOBE_REMOVED = 'side_lobe_removed'

    def __init__(self, srf: AltiusUVSrf, snr=None, normalize_srf=False, solar_distance_AU=1.0):

        self.srf = srf
        solar_spectrum = SolarSpectrum(name='SAO2010')
        wavelength_grid = srf.get_wavelength_grid()

        sao = solar_spectrum.irradiance(wavelength_grid, solardistance=solar_distance_AU)
        solar = xr.DataArray(sao, dims={'wavelength'}, coords={'wavelength': wavelength_grid})

        self._tunable_wavelengths = srf.tunable_wavelengths()

        # fig = plt.figure()
        # camera = Camera(fig)

        blurred_spectrum, central_lobe_portion_srf, central_lobe_fwhm = srf.convolve_full_srf(wavelength_grid, sao,
                                                                                              self._tunable_wavelengths)

        if snr is not None:
            noise = np.random.normal(0, np.sqrt(blurred_spectrum))
            scale = (np.sum(blurred_spectrum) / np.sum(np.abs(noise))) / snr
            noise *= scale
            blurred_spectrum += noise

        self._blurred_spectrum_full_srf = xr.DataArray(blurred_spectrum, dims={'wavelength'},
                                                       coords={'wavelength': self._tunable_wavelengths})

        convolved_side_lobes_removed, central_lobe_portion, central_lobe_fwhm, srf_central_lobe_sum, srf_side_lobe_sum = \
            srf.convolve_corrected_srf(wavelength_grid, sao, self._tunable_wavelengths, self._blurred_spectrum_full_srf)

        central_lobe_portion_gs = srf.get_gaussian_contribution(wavelength_grid, sao, self._tunable_wavelengths)

        self._side_lobes_removed = xr.DataArray(convolved_side_lobes_removed, dims={'wavelength'},
                                                coords={'wavelength': self._tunable_wavelengths})

        self._central_lobe_portions_srf = xr.DataArray(central_lobe_portion_srf, dims={'wavelength'},
                                                       coords={'wavelength': self._tunable_wavelengths})

        self._central_lobe_portions_gauss = xr.DataArray(central_lobe_portion_gs, dims={'wavelength'},
                                                         coords={'wavelength': self._tunable_wavelengths})

        self._central_lobe_fwhm = xr.DataArray(central_lobe_fwhm, dims={'wavelength'},
                                               coords={'wavelength': self._tunable_wavelengths})

        # Weird exception I can't figure out
        try:
            self._srf_central_lobe_sum = xr.DataArray(srf_central_lobe_sum, dims={'convolved_with_solar', 'wavelength'},
                                                      coords={'convolved_with_solar': np.array([False, True]),
                                                              'wavelength': self._tunable_wavelengths})
        except ValueError:
            self._srf_central_lobe_sum = xr.DataArray(srf_central_lobe_sum.T,
                                                      dims={'convolved_with_solar', 'wavelength'},
                                                      coords={'convolved_with_solar': np.array([False, True]),
                                                              'wavelength': self._tunable_wavelengths})

        try:
            self._srf_side_lobe_sum = xr.DataArray(srf_side_lobe_sum, dims={'wavelength', 'convolved_with_solar'},
                                                   coords={'wavelength': self._tunable_wavelengths,
                                                           'convolved_with_solar': np.array([False, True])})
        except ValueError:
            self._srf_side_lobe_sum = xr.DataArray(srf_side_lobe_sum.T, dims={'wavelength', 'convolved_with_solar'},
                                                   coords={'wavelength': self._tunable_wavelengths,
                                                           'convolved_with_solar': np.array([False, True])})
        self._solar = solar

        # self.convolution_progression = camera

    def blurred_spectrum(self, mode, resolution=None, wavelength_shift=0.0, wavelengths=None) -> xr.DataArray:

        xr_data = None
        if mode is self.FULL_SRF:
            xr_data = self._blurred_spectrum_full_srf
        elif mode is self.SIDE_LOBE_REMOVED:
            xr_data = self._side_lobes_removed
        else:
            raise RuntimeError

        if wavelengths is not None:
            interp_val = np.interp(wavelengths, xr_data.wavelength.values, xr_data.values)
            xr_data = xr.DataArray(interp_val, dims={'wavelength'},
                                   coords={'wavelength': wavelengths})

        if resolution is None:
            return xr.DataArray(xr_data.values, dims={'wavelength'},
                                coords={'wavelength': xr_data.wavelength.values - wavelength_shift})
        else:
            # interpolate to specified resolution using the starting wavelength as a start
            high_res_wl = xr_data.wavelength.values
            wl = np.arange(start=high_res_wl[0], stop=high_res_wl[-1], step=resolution)
            interp_spec = np.interp(wl, high_res_wl, xr_data.values)

            return xr.DataArray(interp_spec, dims={'wavelength'},
                                coords={'wavelength': wl - wavelength_shift})

    def spectrum_side_lobes_removed(self):
        return self._side_lobes_removed

    def central_lobe_portions(self) -> (xr.DataArray, xr.DataArray):
        return self._central_lobe_portions_srf, self._central_lobe_portions_gauss

    def central_lobe_fwhm(self):
        return self._central_lobe_fwhm

    def tunable_wavelengths(self):
        return self._tunable_wavelengths

    def unconvolved_solar_spectrum(self):
        return self._solar

    def blur_shift_solar_gaussian(self, fwhm, shift, wavelengths):

        sigma = fwhm_2_sigma(fwhm)
        interp_spec = np.interp(wavelengths, self._solar.wavelength.values, self._solar.values)
        res = wavelengths[1] - wavelengths[0]
        span = 0.4 * (np.max(wavelengths) - np.min(wavelengths))
        mu = np.linspace(-span, span, int(2 * span / res))
        blur = 1 / np.sqrt(2 * np.pi * sigma ** 2) * np.exp(-0.5 * mu ** 2 / sigma ** 2)
        blur /= np.sum(blur)
        blurred_spectrum = xr.DataArray(np.convolve(interp_spec, blur, mode='same'), dims={'wavelength'},
                                        coords={'wavelength': wavelengths - shift})
        return blurred_spectrum, mu, blur

    def srf_central_lobe_sum(self):
        return self._srf_central_lobe_sum

    def srf_side_lobe_sum(self):
        return self._srf_side_lobe_sum


def solar_model(sig, mu, wv, p, solar):
    shifted_wavelen = wv - mu
    solar_spectrum = solar(sig, shifted_wavelen)
    return np.log(solar_spectrum.squeeze()) - np.log(poly_spectrum(wv, p))


def poly_spectrum(w, p):
    return p[5] + p[4] * w + p[3] * w ** 2 + p[2] * w ** 3 + p[1] * w ** 4 + p[0] * w ** 5


def errfunc(params, wavelen, simulated_measurement_spectrum, p, solar):
    fwhm = params[0]
    shift = 0.0
    if len(params) > 1:
        shift = params[1]

    model_guess = solar_model(fwhm, shift, wavelen, p, solar)
    differential_simulated_spectrum = (np.log(simulated_measurement_spectrum) - np.log(poly_spectrum(wavelen, p)))

    # plt.figure()
    # plt.plot(wavelen-shift, model_guess)
    # plt.plot(wavelen, differential_simulated_spectrum, linestyle='dashed')
    return model_guess - differential_simulated_spectrum


def run_fit_to_ideal_no_shift(num_points, FIT_WINDOW=10):
    solar = SAO2010(sigma=fwhm_2_sigma(np.linspace(0.5, 5.0, 46)), start_wavelen=202, end_wavelen=500)

    # Initial guesses:
    init_sigma = fwhm_2_sigma(2.0)
    actual_sigma = fwhm_2_sigma(3.0)

    WL = np.linspace(250, 350, num_points)
    fit_fwhm = np.zeros(len(WL))

    fit_wl = np.copy(WL)

    for k, center_wavelength in enumerate(fit_wl):
        params = np.r_[init_sigma]
        idx = (WL >= (center_wavelength - FIT_WINDOW)) & \
              (WL <= (center_wavelength + FIT_WINDOW))
        assumed_wavelengths = WL[idx]
        meas_slice = solar(actual_sigma, assumed_wavelengths).flatten()
        p = np.polyfit(assumed_wavelengths, meas_slice, 5)
        fit_results = least_squares(
            fun=errfunc, x0=params, loss='cauchy', f_scale=0.03,
            x_scale='jac', verbose=0,
            args=(assumed_wavelengths, meas_slice, p, solar))

        fit_fwhm[k] = sigma_2_fwhm(fit_results.x[0])
        # fit_shift.append(fit_results.x[1])

    return fit_fwhm, sigma_2_fwhm(actual_sigma), fit_wl


def run_fit_to_ideal(num_points, FIT_WINDOW=10):
    solar = SAO2010(sigma=fwhm_2_sigma(np.linspace(0.5, 5.0, 46)), start_wavelen=202, end_wavelen=500)

    # Initial guesses:
    init_sigma = fwhm_2_sigma(2.0)
    init_mu = 0.0
    actual_sigma = fwhm_2_sigma(3.0)
    actual_shift = 0.871

    WL = np.linspace(250, 350, num_points)
    fit_fwhm = np.zeros(len(WL))
    fit_shift = np.zeros(len(WL))

    for k, center_wavelength in enumerate(WL):
        params = np.r_[init_sigma, init_mu]
        idx = (WL >= (center_wavelength - FIT_WINDOW)) & \
              (WL <= (center_wavelength + FIT_WINDOW))
        meas_slice = solar(actual_sigma, WL - actual_shift).flatten()[idx]
        assumed_wavelengths = WL[idx]
        p = np.polyfit(assumed_wavelengths, meas_slice, 5)
        fit_results = least_squares(
            fun=errfunc, x0=params, loss='cauchy', f_scale=0.03,
            x_scale='jac', verbose=0,
            args=(assumed_wavelengths, meas_slice, p, solar))

        fit_fwhm[k] = sigma_2_fwhm(fit_results.x[0])
        fit_shift[k] = fit_results.x[1]

    return fit_fwhm, sigma_2_fwhm(actual_sigma), fit_shift, WL, actual_shift


def run_least_squares_fit_to_data(data_directory, mode, FIT_WINDOW=10, NUM_POINTS=1001):
    SRF = AltiusUVSrf(data_directory)
    TARGET_SHIFT = -0.714

    # Initial guesses:
    init_sigma = fwhm_2_sigma(3.0)
    init_mu = 0.0

    WL = np.linspace(250, 350, NUM_POINTS)

    solar = SAO2010(sigma=fwhm_2_sigma(np.linspace(0.5, 5.0, 46)), start_wavelen=202, end_wavelen=500)

    resolution = 0.01  # nm
    sim_spectrum = SimulatedUVSolarSpectrum(
        SRF)  # trying to fit this with a shifted and blurred solar spectrum with only Gaussian

    fit_fwhm = np.zeros(len(WL))
    fit_shift = []

    for k, center_wavelength in enumerate(WL):
        params = np.r_[init_sigma]
        meas_slice = sim_spectrum.blurred_spectrum(mode=mode, resolution=resolution).sel(
            wavelength=slice(center_wavelength - FIT_WINDOW, center_wavelength + FIT_WINDOW))
        assumed_wavelengths = meas_slice.wavelength.values
        fit_results = fit_measured_spectrum_window_least_squares(assumed_wavelengths,
                                                                 meas_slice.values,
                                                                 params,
                                                                 solar)

        fit_fwhm[k] = sigma_2_fwhm(fit_results.x[0])
        # fit_shift.append(fit_results.x[1])

    actual_fwhm = sim_spectrum.central_lobe_fwhm()

    return fit_fwhm, actual_fwhm, fit_shift, WL, sim_spectrum


def fit_measured_spectrum_window_least_squares(wavelengths: np.ndarray,
                                               measured_spectrum: np.ndarray,
                                               params: np.r_,
                                               solar_model: SAO2010):
    p = np.polyfit(wavelengths, measured_spectrum, 5)
    fit_results = least_squares(
        fun=errfunc, x0=params, loss='cauchy', f_scale=0.03,
        x_scale='jac', verbose=0,
        args=(wavelengths, measured_spectrum, p, solar_model))
    return fit_results


def fit_using_gaussian_approximation(srf: AltiusUVSrf, analysis_wavelengths,
                                     assume_gaussian=True,
                                     snr=None, solar_distance_AU=1.0):
    # TARGET_SHIFT = -0.714

    # Initial guesses:
    init_sigma = fwhm_2_sigma(3.0)
    init_mu = 0.0

    WL = analysis_wavelengths

    sim_spectrum = SimulatedUVSolarSpectrum(srf=srf, snr=snr,
                                            solar_distance_AU=solar_distance_AU)  # trying to fit this with a shifted and blurred solar spectrum with only Gaussian

    fit_fwhm = np.zeros(len(WL))
    rel_percent_diff = np.zeros(len(WL))
    fit_spectrum = np.zeros(len(WL))

    bar = progressbar.ProgressBar(min_value=0, max_value=len(WL))

    fwhm_perturb = FwhmPerturbations(srf=sim_spectrum.srf, solar_distance_AU=solar_distance_AU)

    fwhm_step_function = fwhm_perturb.step_fwhm_assume_gaussian
    if not assume_gaussian:
        fwhm_step_function = fwhm_perturb.step_fwhm_assume_constant_srf_central_lobe_shape

    for k, center_wavelength in enumerate(WL):
        corrected_val = sim_spectrum.blurred_spectrum(sim_spectrum.SIDE_LOBE_REMOVED) \
            .interp(wavelength=center_wavelength).values
        delta_spectrum, fwhm_guess, rel_spectral_delta = fit_fwhm_to_corrected_spectrum(center_wavelength,
                                                                                        corrected_val,
                                                                                        fwhm_step_function)
        bar.update(k)
        rel_percent_diff[k] = 100 * rel_spectral_delta
        fit_spectrum[k] = corrected_val - delta_spectrum
        fit_fwhm[k] = fwhm_guess

    bar.finish()
    actual_fwhm = sim_spectrum.central_lobe_fwhm()

    return fit_fwhm, actual_fwhm, WL.flatten(), sim_spectrum, fit_spectrum, rel_percent_diff


def fit_fwhm_to_corrected_spectrum(center_wavelength, corrected_val, fwhm_step_function):
    fwhm_guess = 3.0
    rel_spectral_delta = 100
    fwhm_rel_delta = 100
    num_iter = 0
    delta_spectrum = None
    deltas = []  # debugging
    while np.abs(rel_spectral_delta) > 0.0005 and num_iter < 800 and fwhm_rel_delta > 0.000001:
        fwhm_next, delta_spectrum = fwhm_step_function(fwhm_guess, center_wavelength, corrected_val, delta_spectrum)
        rel_spectral_delta = delta_spectrum / corrected_val
        fwhm_rel_delta = np.abs(fwhm_next - fwhm_guess) / fwhm_guess
        fwhm_guess = fwhm_next
        num_iter += 1
        deltas.append(delta_spectrum)
    return delta_spectrum, fwhm_guess, rel_spectral_delta


def make_gaussian(center_wavelength, fwhm, wavelengths):
    sigma = fwhm_2_sigma(fwhm)
    g = np.exp(-0.5 * (wavelengths - center_wavelength) ** 2 / sigma ** 2)
    g /= np.sum(g)
    return g
