import itertools

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.patches import Circle
from matplotlib.pyplot import cm

from altius_l1pp_inflight.scene.disk_geometry import center_solar_disk, D_1AU, SolarGeometry
from altius_l1pp_inflight.scene.generation.solar.hmi_files import least_squares_2d  # , HmiSunspotScenario
# from altius_l1pp_inflight.scene.generation.solar.virtual_solar_detector
# import SolarImage, RasterScanGrid, VirtualDetectorPixelArray
from altius_l1pp_inflight.scene.limbdark_mu_pierceandslaughter1977 import LimbDarkening_PS1977


def limb_darkening_test():
    radius = 207
    center_x = 405
    center_y = 1104

    solar_geometry = SolarGeometry(1504, 1504, 752, 752, 0.6, 0.6)
    solar_geometry.set_disk(radius, center_x, center_y, D_1AU)
    mu_ideal = solar_geometry._mu()

    image = create_ideal_solar_image(1504, 1504, center_x, center_y, radius, 335)

    image = image / np.max(image)

    calc_radius, centroid_x, centroid_y = center_solar_disk(image)

    solar_geometry.set_disk(calc_radius, centroid_x, centroid_y, D_1AU)
    mu_calc = solar_geometry._mu()

    limb_darkening = LimbDarkening_PS1977()
    mu_profile = np.linspace(0, 1.0, 50)
    ld_wl_profile = limb_darkening.limb_darkening(335, mu_profile)
    ld_ideal = np.interp(mu_ideal, mu_profile, ld_wl_profile)
    ld_calc = np.interp(mu_calc, mu_profile, ld_wl_profile)
    flattened_ideal = image / ld_ideal
    flattened_calc = image / ld_calc

    x1 = int(center_x - radius / 2)
    x2 = int(center_x + radius / 2)
    y1 = int(center_y - radius / 2)
    y2 = int(center_y + radius / 2)
    coeff, r, rank, s, fitfunc_ideal = least_squares_2d(flattened_ideal[y1:y2, x1:x2], 'TEST_IDEAL')

    coeff, r, rank, s, fitfunc_calc = least_squares_2d(flattened_calc[y1:y2, x1:x2], 'TEST_CALC')

    plt.figure()
    plt.title('Test Solar Image, SNR 1000')
    plt.imshow(np.fliplr(np.flipud(image)))
    plt.plot(center_y, center_x, 'r.')
    plt.plot(centroid_y, centroid_x, 'b+')
    plt.legend(['Ideal', 'Computed'])
    plt.colorbar()

    plt.figure()
    row_idx = int(len(fitfunc_ideal) / 2)
    plt.plot(fitfunc_ideal[row_idx, :], 'r')
    plt.plot(fitfunc_calc[row_idx, :], 'b')
    plt.legend(['Ideal', 'Computed'])

    plt.figure()
    plt.plot(flattened_ideal[y1 + row_idx, x1:x2], 'r', alpha=0.6)
    plt.plot(flattened_calc[y1 + row_idx, x1:x2], 'b', alpha=0.6)
    plt.legend(['Ideal', 'Computed'])
    plt.show()


def test_limb_darkening_interpolation():
    limb_darkening = LimbDarkening_PS1977()
    mu_profile = np.linspace(0, 1.0, 50)
    ld_wl_profile_1 = limb_darkening.limb_darkening(610.975, mu_profile)
    ld_wl_profile_2 = limb_darkening.limb_darkening(620.590, mu_profile)
    ld_wl = limb_darkening.limb_darkening(617.3, mu_profile)

    plt.plot(mu_profile, ld_wl_profile_1)
    plt.plot(mu_profile, ld_wl_profile_2)
    plt.plot(mu_profile, ld_wl)
    plt.legend(['610.975', '620.59', '617.3'])

    solar_geometry = SolarGeometry(1504, 1504, 752, 740, 0.6, 0.6)
    solar_geometry.set_disk(200, 350, 1010, D_1AU)

    mu_ideal = solar_geometry._mu()

    ld_ideal = np.interp(mu_ideal, mu_profile, ld_wl)


def plot_centering():
    radius = 207.5
    center_h = 405.3
    center_v = 175  # 1104.15

    image_raw = create_ideal_solar_image(1503, 1503, center_h, center_v, radius, 335)
    calc_radius_raw, x_c_raw, y_c_raw, alg_raw = center_solar_disk(image_raw,
                                                                   **{'detector_angular_width_degrees': (2.0, 2.0),
                                                                      'distance_to_sun_km': D_1AU})

    image_sci = create_ideal_solar_image(501, 501, center_h / 3, center_v / 3, radius / 3, 335, binning=9)
    calc_radius_sci, x_c_sci, y_c_sci, alg_sci = center_solar_disk(image_sci,
                                                                   **{'detector_angular_width_degrees': (2.0, 2.0),
                                                                      'distance_to_sun_km': D_1AU})

    fig, ax = plt.subplots()
    ax.imshow(image_raw, cmap='Greys')
    legend = []

    color = iter(cm.rainbow(np.linspace(0, 1, len(alg_raw['found_radii']))))
    marker = itertools.cycle((',', '+', '.', 'o', '*', 'x'))
    for k in range(0, len(alg_raw['found_radii'])):
        h_edge = alg_raw['horizontal_edge_points'][k]  # three points
        v_edge = alg_raw['vertical_edge_points'][k]  # three points
        r_iter = alg_raw['found_radii'][k]
        found_h_c = alg_raw['horizontal_center'][k]
        found_v_c = alg_raw['vertical_center'][k]
        c = next(color)
        m = next(marker)
        ax.plot(np.append(h_edge, found_h_c), np.append(v_edge, found_v_c),
                marker=m, markersize=7, color=c, linestyle='None')
        legend.append('Edge iter {}'.format(k))
        # ax.plot(found_h_c, found_v_c, marker=m, color=c)
        # ax.plot(x_c_raw, y_c_raw, 'r.')

    ax.plot(x_c_raw, y_c_raw, 'bo', markersize=8)
    legend.append('Found center')
    ax.plot(center_h, center_v, 'ro', markersize=8)
    legend.append('True center')
    # draw a circle using the radius and center found:
    circ = Circle((center_h, center_v), radius, fill=False, color='r')
    ax.add_patch(circ)
    ax.legend(legend)
    plt.title('Solar Centering at Detector Resolution, Partial Disk')

    fig, ax = plt.subplots()
    ax.imshow(image_sci, cmap='Greys')
    legend = []

    color = iter(cm.rainbow(np.linspace(0, 1, len(alg_sci['found_radii']))))
    marker = itertools.cycle((',', '+', '.', 'o', '*', 'x'))
    for k in range(0, len(alg_sci['found_radii'])):
        h_edge = alg_sci['horizontal_edge_points'][k]  # three points
        v_edge = alg_sci['vertical_edge_points'][k]  # three points
        r_iter = alg_sci['found_radii'][k]
        found_h_c = alg_sci['horizontal_center'][k]
        found_v_c = alg_sci['vertical_center'][k]
        c = next(color)
        m = next(marker)
        ax.plot(np.append(h_edge, found_h_c), np.append(v_edge, found_v_c),
                marker=m, markersize=7, linestyle='None', color=c)
        legend.append('Edge iter {}'.format(k))
        # ax.plot(x_c_raw, y_c_raw, 'r.')

    ax.plot(x_c_sci, y_c_sci, 'bo', markersize=8)
    legend.append('Found center')
    ax.plot(center_h / 3, center_v / 3, 'ro', markersize=8)
    legend.append('True center')

    # draw a circle using the radius and center found:
    circ = Circle((center_h / 3, center_v / 3), radius / 3, fill=False, color='r')
    ax.add_patch(circ)
    ax.legend(legend)
    plt.title('Solar Centering at Science Resolution, Partial Disk')
    plt.show()


# def centroid_and_limb_correct_hmi():
#     # HMI:
#     xw = 4096
#     yw = 4096
#
#     solar_image = SolarImage.from_netcdf(HmiSunspotScenario.NO_SUNSPOTS, 8, 8, 30)
#     radius = solar_image._dataset['disk_radius_pixels'].values
#     # solar_image.visualize_detector_image()
#
#     image = solar_image.hmi_disk_radiance()
#     dh, dv = np.meshgrid(np.arange(0, 4096), np.arange(0, 4096))
#
#     xbar = np.nansum(dh * image) / np.nansum(image)
#     ybar = np.nansum(dv * image) / np.nansum(image)
#
#     print('Centroided HMI image center: ', xbar, ', ', ybar)
#
#     # image = create_ideal_solar_image(xw, yw, center_x, center_y, radius, 335, 100)
#     # plt.imshow(image)
#     # plt.colorbar()
#     # plt.show()
#
#     calc_radius, x_c, y_c = center_solar_disk(image)
#     print('calc_radius_hmi: ', calc_radius)
#     print('x_c_hmi: ', x_c)
#     print('y_c_hmi: ', y_c)
#
#     # need better than 1 pixel at ALTIUS resolution
#     # assert np.isclose(calc_radius, radius, atol=4.0), 'Calculated Radius: ' + str(calc_radius) + ',
#     Actual Radius: ' + str(radius)
#     assert np.isclose(x_c, xbar, atol=4.0), 'Calculated x_c: ' + str(x_c) + ', Actual x_c: ' + str(xbar)
#     assert np.isclose(y_c, ybar, atol=4.0), 'Calculated y_c: ' + str(y_c) + ', Actual x_c: ' + str(ybar)
#
#     solar_geometry = SolarGeometry(1504, 1504, 752, 752)
#     solar_geometry.set_disk(radius, xbar, ybar, D_1AU)
#     mu_ideal = solar_geometry._mu()
#
#     solar_geometry.set_disk(calc_radius, x_c, y_c, D_1AU)
#     mu_calc = solar_geometry._mu()
#
#     limb_darkening = LimbDarkening_PS1977()
#     mu_profile = np.linspace(0, 1.0, 50)
#     ld_wl_profile = limb_darkening.limb_darkening(617.3, mu_profile)
#     ld_ideal = np.interp(mu_ideal, mu_profile, ld_wl_profile)
#     ld_calc = np.interp(mu_calc, mu_profile, ld_wl_profile)
#     flattened_ideal = image / ld_ideal
#     flattened_calc = image / ld_calc
#
#     x1 = int(xbar - radius / 2)
#     x2 = int(xbar + radius / 2)
#     y1 = int(ybar - radius / 2)
#     y2 = int(ybar + radius / 2)
#     coeff, r, rank, s, fitfunc_ideal = least_squares_2d(flattened_ideal[y1:y2, x1:x2], 'TEST_IDEAL')
#
#     coeff, r, rank, s, fitfunc_calc = least_squares_2d(flattened_calc[y1:y2, x1:x2], 'TEST_CALC')
#
#     fig, ax = plt.subplots()
#     im = ax.imshow(image)
#     ax.plot(ybar, xbar, 'b.')
#     ax.plot(y_c, x_c, 'r+')
#     plt.legend(['Ideal', 'Computed'])
#     fig.colorbar(im)
#     # draw a circle using the radius and center found:
#     circ = Circle((x_c, y_c), calc_radius, fill=False, color='r')
#     ax.add_patch(circ)
#     plt.show()
#
#     plt.figure()
#     row_idx = int(len(fitfunc_ideal) / 2)
#     plt.plot(fitfunc_ideal[row_idx, :], 'r')
#     plt.plot(fitfunc_calc[row_idx, :], 'b')
#     plt.legend(['Ideal', 'Computed'])
#
#     plt.figure()
#     plt.plot(flattened_ideal[y1 + row_idx, x1:x2], 'r', alpha=0.6)
#     plt.plot(flattened_calc[y1 + row_idx, x1:x2], 'b', alpha=0.6)
#     plt.legend(['Ideal', 'Computed'])
#
#     plt.figure()
#     plt.imshow(flattened_calc[y1:y2, x1:x2])
#     plt.colorbar()
#     plt.show()


def center_disk_partially_off_detector():
    print('*********************************')
    print('Performing ideal ALTIUS-resolution (1504x1504) solar +scene image')
    print(' with sun partially off detector')
    print('Tolerance is expected to be within 3.0 pixels')
    print('*********************************')
    print()
    radius = 207
    center_x = 405
    center_y = 1455

    # ALTIUS:
    image = create_ideal_solar_image(1504, 1504, center_x, center_y, radius, 335)

    plt.imshow(image)
    plt.colorbar()
    plt.show()

    calc_radius, x_c, y_c = center_solar_disk(image, **{'detector_angular_width_degrees': (2.0, 2.0),
                                                        'distance_to_sun_km': 147.24e6})

    fig, ax = plt.subplots()
    ax.imshow(image)
    ax.plot(x_c, y_c, 'r.')
    # draw a circle using the radius and center found:
    circ = Circle((x_c, y_c), calc_radius, fill=False, color='r')
    ax.add_patch(circ)
    plt.show()

    # need better than 1 pixel at ALTIUS resolution
    assert np.isclose(calc_radius, radius, atol=3.0), 'Radius off by ' + str(np.abs(calc_radius - radius)) + ' pixels'
    assert np.isclose(x_c, center_x, atol=3.0), 'X center off by ' + str(np.abs(x_c - center_x)) + ' pixels'
    assert np.isclose(y_c, center_y, atol=3.0), 'Y center off by ' + str(np.abs(y_c - center_y)) + ' pixels'

    print('Expected radius: ', radius)
    print('Actual radius: ', calc_radius)
    print()
    print('Expected horizontal center: ', center_x)
    print('Actual horizontal center: ', x_c)
    print()
    print('Expected vertical center: ', center_y)
    print('Actual vertical center: ', y_c)
    print()
    print('*********************************')
    print('Centroid test off-detector passed')
    print('*********************************')


def centroid_arbitrary_disk():
    print('*********************************')
    print('Performing ideal ALTIUS-resolution (1504x1504) solar +scene image')
    print('Tolerance is expected to be within 3.0 pixels')
    print('*********************************')
    print()
    radius = 207
    center_x = 405
    center_y = 1104

    # ALTIUS:
    image = create_ideal_solar_image(1504, 1504, center_x, center_y, radius, 335)

    plt.imshow(image)
    plt.colorbar()
    plt.show()

    calc_radius, x_c, y_c = center_solar_disk(image)

    fig, ax = plt.subplots()
    ax.imshow(image)
    ax.plot(x_c, y_c, 'r.')
    # draw a circle using the radius and center found:
    circ = Circle((x_c, y_c), calc_radius, fill=False, color='r')
    ax.add_patch(circ)
    plt.show()

    # need better than 1 pixel at ALTIUS resolution
    assert np.isclose(calc_radius, radius, atol=3.0)
    assert np.isclose(x_c, center_x, atol=3.0)
    assert np.isclose(y_c, center_y, atol=3.0)

    print('Expected radius: ', radius)
    print('Actual radius: ', calc_radius)
    print()
    print('Expected horizontal center: ', center_x)
    print('Actual horizontal center: ', x_c)
    print()
    print('Expected vertical center: ', center_y)
    print('Actual vertical center: ', y_c)
    print()
    print('*********************************')
    print('Centroid test passed')
    print('*********************************')


# def centroid_altius_disk():
#     xw = 4096
#     yw = 4096
#
#     solar_image = SolarImage.from_netcdf(HmiSunspotScenario.NO_SUNSPOTS, 8, 8, 30)
#     radius = solar_image._dataset['disk_radius_pixels'].values
#
#     rg = RasterScanGrid(solar_radius_angular_size_deg=solar_image._dataset['disk_radius_apparent_angle'].values,
#                         scan_rate_deg_per_sec=1.0,
#                         time_per_image_sec=1.0,
#                         fov_horizontal_deg=solar_image._dataset['fov_horizontal_deg'].values,
#                         fov_vertical_deg=solar_image._dataset['fov_vertical_deg'].values,
#                         v_in=solar_image._dataset['num_vertical_scans'].values,
#                         h_in=solar_image._dataset['num_horizontal_scans'].values)
#
#     vp = VirtualDetectorPixelArray(disk_radius_pixels=radius,
#                                    solar_scan_grid=rg,
#                                    boresight_horz_portion=0.0,
#                                    boresight_vert_portion=0.0)
#
#     vp.set_current_raster_image(solar_image, 1504, 1504, 35)
#     altius_image, bbox = vp.to_altius_resolution_image()
#
#     radius, horz_center, vert_center = center_solar_disk(altius_image)
#
#     dh, dv = np.meshgrid(np.arange(0, 1504), np.arange(0, 1504))
#     xbar = np.nansum(dh * altius_image) / np.nansum(altius_image)
#     ybar = np.nansum(dv * altius_image) / np.nansum(altius_image)
#
#     fig, ax = plt.subplots()
#     ax.imshow(altius_image)
#     ax.plot(horz_center, vert_center, 'r.')
#     ax.plot(xbar, ybar, 'b+')
#     # draw a circle using the radius and center found:
#     circ = Circle((horz_center, vert_center), radius, fill=False, color='r')
#     ax.add_patch(circ)
#
#     solar_geometry = SolarGeometry(1504, 1504, 752, 752)
#     solar_geometry.set_disk(radius, horz_center, vert_center, D_1AU)
#
#     mu_calc = solar_geometry._mu()
#
#     limb_darkening = LimbDarkening_PS1977()
#     mu_profile = np.linspace(0, 1.0, 50)
#     ld_wl_profile = limb_darkening.limb_darkening(617.3, mu_profile)
#     ld_calc = np.interp(mu_calc, mu_profile, ld_wl_profile)
#     # flattened_calc = image / ld_calc
#     #
#     # x1 = int(horz_center - radius / 2)
#     # x2 = int(horz_center + radius / 2)
#     # y1 = int(vert_center - radius / 2)
#     # y2 = int(vert_center + radius / 2)
#
#     plt.figure()
#     im_profile = altius_image[int(vert_center), int(horz_center - radius):int(horz_center + radius)]
#     ld_p = ld_calc[int(vert_center), int(horz_center - radius):int(horz_center + radius)]
#     plt.plot(im_profile / np.nanmax(im_profile))
#     plt.plot(ld_p)
#
#     plt.show()


def create_ideal_solar_image(hsize, vsize, h_center, v_center, radius, wavelen, snr=1000, binning=1):
    # create a disk, apply limb darkening, then add reasonable noise
    high_dn = 14000 * binning
    image = np.zeros((vsize, hsize))

    X, Y = np.meshgrid(np.arange(0, hsize), np.arange(0, vsize))
    R = np.sqrt((X - h_center) ** 2 + (Y - v_center) ** 2)
    mu = np.sqrt(1 - (R / radius) ** 2)
    solar_disk_idx = R <= radius
    image[solar_disk_idx] = high_dn
    darkening_model = LimbDarkening_PS1977()
    limb_profile = darkening_model.limb_darkening(wavelen, np.linspace(0, 1, 100))
    scale = np.interp(mu, np.linspace(0, 1, 100), limb_profile)
    image *= scale
    image[R > radius] = high_dn / snr
    image = np.random.poisson(image * 2, np.shape(image)) / 2 / binning
    return image


if __name__ == '__main__':
    centroid_arbitrary_disk()
    center_disk_partially_off_detector()
    limb_darkening_test()
    # centroid_and_limb_correct_hmi()
    # test_limb_darkening_interpolation()
    # centroid_altius_disk()
